from __future__ import annotations

import inspect
import typing

from inngest._internal import (
    errors,
    event_lib,
    execution,
    function,
    transforms,
    types,
)

from .log import LoggerMiddleware
from .middleware import Middleware, MiddlewareSync, UninitializedMiddleware

if typing.TYPE_CHECKING:
    from inngest._internal import client_lib

_mismatched_sync = errors.AsyncUnsupportedError(
    "encountered async middleware in non-async context"
)

DEFAULT_CLIENT_MIDDLEWARE: list[UninitializedMiddleware] = [LoggerMiddleware]


class MiddlewareManager:
    @property
    def middleware(self) -> list[typing.Union[Middleware, MiddlewareSync]]:
        return [*self._middleware]

    def __init__(self, client: client_lib.Inngest) -> None:
        self.client = client
        self._disabled_hooks = set[str]()
        self._middleware = list[typing.Union[Middleware, MiddlewareSync]]()

    @classmethod
    def from_client(cls, client: client_lib.Inngest) -> MiddlewareManager:
        """
        Create a new manager from an Inngest client, using the middleware on the
        client.
        """
        mgr = cls(client)

        for m in DEFAULT_CLIENT_MIDDLEWARE:
            mgr.add(m)

        for m in client.middleware:
            mgr.add(m)

        return mgr

    @classmethod
    def from_manager(cls, manager: MiddlewareManager) -> MiddlewareManager:
        """
        Create a new manager from another manager, using the middleware on the
        passed manager. Effectively wraps a manager.
        """
        new_mgr = cls(manager.client)
        for m in manager.middleware:
            new_mgr._middleware = [*new_mgr._middleware, m]
        return new_mgr

    def add(self, middleware: UninitializedMiddleware) -> None:
        self._middleware = [*self._middleware, middleware(self.client)]

    async def after_execution(self) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                await transforms.maybe_await(m.after_execution())
            return None
        except Exception as err:
            return err

    def after_execution_sync(self) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.after_execution):
                    return _mismatched_sync
                m.after_execution()
            return None
        except Exception as err:
            return err

    async def before_execution(self) -> types.MaybeError[None]:
        hook = "before_execution"
        if hook in self._disabled_hooks:
            # Only allow before_execution to be called once. This simplifies
            # code since execution can start at the function or step level.
            return None

        try:
            for m in self._middleware:
                await transforms.maybe_await(m.before_execution())

            self._disabled_hooks.add(hook)
            return None
        except Exception as err:
            return err

    def before_execution_sync(self) -> types.MaybeError[None]:
        hook = "before_execution"
        if hook in self._disabled_hooks:
            # Only allow before_execution to be called once. This simplifies
            # code since execution can start at the function or step level.
            return None

        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.before_execution):
                    return _mismatched_sync
                m.before_execution()

            self._disabled_hooks.add(hook)
            return None
        except Exception as err:
            return err

    async def before_response(self) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                await transforms.maybe_await(m.before_response())
            return None
        except Exception as err:
            return err

    def before_response_sync(self) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.before_response):
                    return _mismatched_sync
                m.before_response()
            return None
        except Exception as err:
            return err

    async def before_send_events(
        self,
        events: list[event_lib.Event],
    ) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                await transforms.maybe_await(m.before_send_events(events))
            return None
        except Exception as err:
            return err

    def before_send_events_sync(
        self,
        events: list[event_lib.Event],
    ) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.before_send_events):
                    return _mismatched_sync
                m.before_send_events(events)
            return None
        except Exception as err:
            return err

    async def transform_input(
        self,
        ctx: function.Context,
    ) -> types.MaybeError[function.Context]:
        try:
            for m in self._middleware:
                ctx = await transforms.maybe_await(
                    m.transform_input(ctx),
                )
            return ctx
        except Exception as err:
            return err

    def transform_input_sync(
        self,
        ctx: function.Context,
    ) -> types.MaybeError[function.Context]:
        try:
            for m in self._middleware:
                if isinstance(m, Middleware):
                    return _mismatched_sync
                ctx = m.transform_input(ctx)
            return ctx
        except Exception as err:
            return err

    async def transform_output(
        self,
        output: typing.Optional[execution.Output],
    ) -> types.MaybeError[typing.Optional[execution.Output]]:
        # Nothing to transform
        if output is None:
            return None

        try:
            for m in self._middleware:
                output = await transforms.maybe_await(
                    m.transform_output(output)
                )
            return output
        except Exception as err:
            return err

    def transform_output_sync(
        self,
        output: typing.Optional[execution.Output],
    ) -> types.MaybeError[typing.Optional[execution.Output]]:
        # Nothing to transform
        if output is None:
            return None

        try:
            for m in self._middleware:
                if isinstance(m, Middleware):
                    return _mismatched_sync
                output = m.transform_output(output)
            return output
        except Exception as err:
            return err



"""
Simulator for the DigitalOcean Function runtime.
"""

import dataclasses
import json
import typing

import flask

_PATH_PREFIX = "/api/v1/web"
_NAMESPACE = "fn-00000000"
_FUNCTION_NAME = "fake-name"
FULL_PATH = f"{_PATH_PREFIX}/{_NAMESPACE}/{_FUNCTION_NAME}"

_Main = typing.Callable[[typing.Any, typing.Any], typing.Any]


@dataclasses.dataclass
class _Context:
    api_host: str
    function_name: str


class DigitalOceanSimulator:
    """
    Simulator for the DigitalOcean Function runtime. This is necessary because
    DigitalOcean Functions don't seem to have a local development environment.
    If that's wrong or changes, this class can likely be deleted.
    """

    def __init__(self, main: _Main) -> None:
        """
        Args:
        ----
            main: DigitalOcean Function handler.
        """
        self.app = _create_app(main)


def _create_app(main: _Main) -> flask.Flask:
    app = flask.Flask(__name__)

    @app.route(FULL_PATH, methods=["GET", "POST", "PUT"])
    def handler() -> flask.Response:
        event: dict[str, object] = {
            "http": {
                "body": flask.request.data.decode("utf-8"),
                "headers": dict(flask.request.headers.items()),
                "method": flask.request.method,
                "path": "",
                "queryString": flask.request.query_string.decode("utf-8"),
            },
        }

        res = main(
            event,
            _Context(
                api_host=flask.request.url_root,
                function_name=f"/{_NAMESPACE}/{_FUNCTION_NAME}",
            ),
        )
        if not isinstance(res, dict):
            raise ValueError("response must be a dict")

        body = None
        if "body" in res:
            body = json.dumps(res["body"])

        headers = res.get("headers", {})
        if not isinstance(headers, dict):
            raise ValueError("headers must be a dict")

        status_code = res.get("statusCode", 200)
        if not isinstance(status_code, int):
            raise ValueError("statusCode must be an int")

        return flask.Response(
            headers=headers,
            response=body,
            status=status_code,
        )

    return app



from __future__ import annotations

from inngest._internal import client_lib, function, types

from .middleware import MiddlewareSync


class LoggerProxy:
    _proxied_methods = (
        "critical",
        "debug",
        "error",
        "exception",
        "fatal",
        "info",
        "log",
        "warn",
        "warning",
    )

    def __init__(self, logger: types.Logger) -> None:
        self._is_enabled = False
        self.logger = logger

    def __getattr__(self, name: str) -> object:
        if name in self._proxied_methods and not self._is_enabled:
            # Return noop
            return lambda *args, **kwargs: None

        return getattr(self.logger, name)

    def enable(self) -> None:
        self._is_enabled = True


class LoggerMiddleware(MiddlewareSync):
    def __init__(self, client: client_lib.Inngest) -> None:
        super().__init__(client)
        self.logger = LoggerProxy(client.logger)

    def before_execution(self) -> None:
        self.logger.enable()

    def transform_input(
        self,
        ctx: function.Context,
    ) -> function.Context:
        self.logger.logger = ctx.logger
        ctx.logger = self.logger  # type: ignore
        return ctx



from __future__ import annotations

import typing

from inngest._internal import event_lib, execution, function

if typing.TYPE_CHECKING:
    from inngest._internal import client_lib


class Middleware:
    def __init__(self, client: client_lib.Inngest) -> None:
        self._client = client

    async def after_execution(self) -> None:
        """
        After executing new code. Called multiple times per run when using
        steps.
        """
        return None

    async def before_execution(self) -> None:
        """
        Before executing new code. Called multiple times per run when using
        steps.
        """
        return None

    async def before_response(self) -> None:
        """
        After the output has been set and before the response is sent
        back to Inngest. This is where you can perform any final actions before
        the response is sent back to Inngest. Called multiple times per run when
        using steps. Not called for function middleware.
        """
        return None

    async def before_send_events(self, events: list[event_lib.Event]) -> None:
        """
        Before sending events.
        """
        return None

    async def transform_input(
        self,
        ctx: function.Context,
    ) -> function.Context:
        """
        Before calling a function or step. Used to replace certain arguments in
        the function. Called multiple times per run when using steps.
        """
        return ctx

    async def transform_output(
        self,
        output: execution.Output,
    ) -> execution.Output:
        """
        After a function or step returns. Used to modify the returned data.
        Called multiple times per run when using steps. Not called when an error
        is thrown.
        """
        return output


class MiddlewareSync:
    client: client_lib.Inngest

    def __init__(self, client: client_lib.Inngest) -> None:
        self.client = client

    def after_execution(self) -> None:
        """
        After executing new code. Called multiple times per run when using
        steps.
        """
        return None

    def before_execution(self) -> None:
        """
        Before executing new code. Called multiple times per run when using
        steps.
        """
        return None

    def before_response(self) -> None:
        """
        After the output has been set and before the response is sent
        back to Inngest. This is where you can perform any final actions before
        the response is sent back to Inngest. Called multiple times per run when
        using steps. Not called for function middleware.
        """
        return None

    def before_send_events(self, events: list[event_lib.Event]) -> None:
        """
        Before sending events.
        """
        return None

    def transform_input(
        self,
        ctx: function.Context,
    ) -> function.Context:
        """
        Before calling a function or step. Used to replace certain arguments in
        the function. Called multiple times per run when using steps.
        """
        return ctx

    def transform_output(
        self,
        output: execution.Output,
    ) -> execution.Output:
        """
        After a function or step returns. Used to modify the returned data.
        Called multiple times per run when using steps. Not called when an error
        is thrown.
        """
        return output


UninitializedMiddleware = typing.Callable[
    # Used a "client_lib.Inngest" string to avoid a circular import
    ["client_lib.Inngest"], typing.Union[Middleware, MiddlewareSync]
]



from __future__ import annotations

import typing

from . import types


class Event(types.BaseModel):
    data: typing.Mapping[str, types.JSON] = {}
    id: str = ""
    name: str
    ts: int = 0
    user: typing.Mapping[str, types.JSON] = {}


# Necessary because of the recursive JSON type
Event.model_rebuild()



import os
import unittest

import pytest

from . import client_lib, const, errors, event_lib


class Test(unittest.TestCase):
    def test_dev_env_var(self) -> None:
        """
        No error is raised when the INNGEST_DEV environment variable is set.
        This env var is a fallback for the is_production param.
        """

        os.environ[const.EnvKey.DEV.value] = "1"
        self.addCleanup(lambda: os.environ.pop(const.EnvKey.DEV.value))

        client_lib.Inngest(
            app_id="test",
        )

    def test_env_env_var(self) -> None:
        """
        If INNGEST_ENV is set, it is used as the env
        """

        os.environ[const.EnvKey.ENV.value] = "my-env"
        self.addCleanup(lambda: os.environ.pop(const.EnvKey.ENV.value))
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )
        assert client.env == "my-env"

    def test_env_param(self) -> None:
        """
        If the env param is set, it takes precedence over the INNGEST_ENV env
        var
        """

        os.environ[const.EnvKey.ENV.value] = "my-env1"
        self.addCleanup(lambda: os.environ.pop(const.EnvKey.ENV.value))
        client = client_lib.Inngest(
            app_id="test",
            env="my-env2",
            signing_key="foo",
        )
        assert client.env == "my-env2"

    def test_event_key_env_var(self) -> None:
        os.environ[const.EnvKey.EVENT_KEY.value] = "foo2"
        self.addCleanup(lambda: os.environ.pop(const.EnvKey.EVENT_KEY.value))
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )
        assert client.event_key == "foo2"

    def test_event_key_param(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            event_key="foo1",
            signing_key="foo",
        )
        assert client.event_key == "foo1"

    def test_event_key_missing(self) -> None:
        """
        Error is raised when the event key is not set in production.
        """

        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )

        with pytest.raises(errors.EventKeyUnspecifiedError):
            client.send_sync(event_lib.Event(name="foo"))

    def test_signing_key_env_var(self) -> None:
        os.environ[const.EnvKey.SIGNING_KEY.value] = "foo2"
        self.addCleanup(lambda: os.environ.pop(const.EnvKey.SIGNING_KEY.value))
        client = client_lib.Inngest(app_id="test")
        assert client.signing_key == "foo2"

    def test_signing_key_param(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo1",
        )
        assert client.signing_key == "foo1"

    def test_cloud_mode_without_signing_key(self) -> None:
        """
        When in Cloud mode but no signing key, no error is raised. This behavior
        is necessary because some users may only use the Inngest client for
        sending events. The signing key should only be required when serving
        Inngest functions in Cloud mode
        """

        client_lib.Inngest(app_id="test")

    def test_api_base_url_env_var(self) -> None:
        os.environ[const.EnvKey.API_BASE_URL.value] = "example.com"
        self.addCleanup(lambda: os.environ.pop(const.EnvKey.API_BASE_URL.value))
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )
        assert client.api_origin == "example.com"

    def test_api_base_url_param(self) -> None:
        client = client_lib.Inngest(
            api_base_url="example.com",
            app_id="test",
            signing_key="foo",
        )
        assert client.api_origin == "example.com"

    def test_api_base_url_default_prod(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )
        assert client.api_origin == "https://api.inngest.com/"

    def test_api_base_url_default_dev(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            is_production=False,
            signing_key="foo",
        )
        assert client.api_origin == "http://127.0.0.1:8288/"

    def test_event_api_base_url_env_var(self) -> None:
        os.environ[const.EnvKey.EVENT_API_BASE_URL.value] = "example.com"
        self.addCleanup(
            lambda: os.environ.pop(const.EnvKey.EVENT_API_BASE_URL.value)
        )
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )
        assert client.event_api_origin == "example.com"

    def test_event_api_base_url_param(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            event_api_base_url="example.com",
            signing_key="foo",
        )
        assert client.event_api_origin == "example.com"

    def test_eventapi_base_url_default_prod(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            signing_key="foo",
        )
        assert client.event_api_origin == "https://inn.gs/"

    def test_eventapi_base_url_default_dev(self) -> None:
        client = client_lib.Inngest(
            app_id="test",
            is_production=False,
            signing_key="foo",
        )
        assert client.event_api_origin == "http://127.0.0.1:8288/"



import os
import typing

from . import const


def get_environment_name() -> typing.Optional[str]:
    """
    Get the Inngest environment name from env vars. Checks our env var and cloud
    platform (e.g. Vercel) specific env vars
    """

    return (
        os.getenv(const.EnvKey.ENV.value)
        or os.getenv(const.EnvKey.RAILWAY_GIT_BRANCH.value)
        or os.getenv(const.EnvKey.RENDER_GIT_BRANCH.value)
        or os.getenv(const.EnvKey.VERCEL_GIT_BRANCH.value)
    )


def is_true(env_var: const.EnvKey) -> bool:
    val = os.getenv(env_var.value)
    if val is None:
        return False

    if val.lower() in ("true", "1"):
        return True

    return False



from __future__ import annotations

import hashlib
import hmac
import http
import os
import threading
import typing
import urllib.parse

import httpx

from . import const, errors, transforms, types


class ThreadAwareAsyncHTTPClient(httpx.AsyncClient):
    """
    Thin wrapper around httpx.AsyncClient. It keeps track of the thread it was
    created in, which is critical since asyncio is not thread safe: calling an
    async method in a different thread will raise an exception
    """

    _creation_thread_id: typing.Optional[int] = None

    def is_same_thread(self) -> bool:
        if self._creation_thread_id is None:
            raise Exception("did initialize ThreadAwareAsyncHTTPClient")

        current_thread_id = threading.get_ident()
        return self._creation_thread_id == current_thread_id

    def initialize(self) -> ThreadAwareAsyncHTTPClient:
        self._creation_thread_id = threading.get_ident()
        return self


def create_headers(
    *,
    env: typing.Optional[str],
    framework: typing.Optional[const.Framework],
    server_kind: typing.Optional[const.ServerKind],
) -> dict[str, str]:
    """
    Create standard headers that should exist on every possible outgoing
    request.
    """

    headers = {
        const.HeaderKey.CONTENT_TYPE.value: "application/json",
        const.HeaderKey.SDK.value: f"inngest-{const.LANGUAGE}:v{const.VERSION}",
        const.HeaderKey.USER_AGENT.value: f"inngest-{const.LANGUAGE}:v{const.VERSION}",
    }

    if env is not None:
        headers[const.HeaderKey.ENV.value] = env
    if framework is not None:
        headers[const.HeaderKey.FRAMEWORK.value] = framework.value
    if server_kind is not None:
        headers[const.HeaderKey.EXPECTED_SERVER_KIND.value] = server_kind.value

    return headers


def create_serve_url(
    *,
    request_url: str,
    serve_origin: typing.Optional[str],
    serve_path: typing.Optional[str],
) -> str:
    """
    Create the serve URL, which is the URL that the Executor will use to reach
    the SDK.

    Args:
    ----
        request_url: The URL that the Executor is using to reach the SDK.
        serve_origin: User-specified override for the serve origin.
        serve_path: User-specified override for the serve path.
    """

    # User can also specify these via env vars. The env vars take precedence.
    serve_origin = os.getenv(const.EnvKey.SERVE_ORIGIN.value, serve_origin)
    serve_path = os.getenv(const.EnvKey.SERVE_PATH.value, serve_path)

    parsed_url = urllib.parse.urlparse(request_url)
    new_scheme = parsed_url.scheme
    new_netloc = parsed_url.netloc
    new_path = parsed_url.path

    if serve_origin is not None:
        has_scheme = "://" in serve_origin
        if has_scheme:
            parsed_origin = urllib.parse.urlparse(serve_origin)
            new_scheme = parsed_origin.scheme or new_scheme
            new_netloc = parsed_origin.netloc or new_netloc
        else:
            new_scheme = "https"
            new_netloc = serve_origin

    if serve_path is not None:
        new_path = serve_path

    return urllib.parse.urlunparse(
        (new_scheme, new_netloc, new_path, "", "", "")
    )


async def fetch_with_auth_fallback(
    logger: types.Logger,
    client: ThreadAwareAsyncHTTPClient,
    client_sync: httpx.Client,
    request: httpx.Request,
    *,
    signing_key: typing.Optional[str],
    signing_key_fallback: typing.Optional[str],
) -> httpx.Response:
    """
    Send an HTTP request with the given signing key. If the response is a 401 or
    403, then try again with the fallback signing key
    """

    if signing_key is not None:
        request.headers[
            const.HeaderKey.AUTHORIZATION.value
        ] = f"Bearer {transforms.hash_signing_key(signing_key)}"

    res = await fetch_with_thready_safety(
        logger,
        client,
        client_sync,
        request,
    )
    if (
        res.status_code
        in (http.HTTPStatus.FORBIDDEN, http.HTTPStatus.UNAUTHORIZED)
        and signing_key_fallback is not None
    ):
        # Try again with the signing key fallback
        request.headers[
            const.HeaderKey.AUTHORIZATION.value
        ] = f"Bearer {transforms.hash_signing_key(signing_key_fallback)}"

        res = await fetch_with_thready_safety(
            logger,
            client,
            client_sync,
            request,
        )

    return res


def fetch_with_auth_fallback_sync(
    client: httpx.Client,
    request: httpx.Request,
    *,
    signing_key: typing.Optional[str],
    signing_key_fallback: typing.Optional[str],
) -> httpx.Response:
    """
    Send an HTTP request with the given signing key. If the response is a 401 or
    403, then try again with the fallback signing key
    """

    if signing_key is not None:
        request.headers[
            const.HeaderKey.AUTHORIZATION.value
        ] = f"Bearer {transforms.hash_signing_key(signing_key)}"

    res = client.send(request)
    if (
        res.status_code
        in (http.HTTPStatus.FORBIDDEN, http.HTTPStatus.UNAUTHORIZED)
        and signing_key_fallback is not None
    ):
        # Try again with the signing key fallback
        request.headers[
            const.HeaderKey.AUTHORIZATION.value
        ] = f"Bearer {transforms.hash_signing_key(signing_key_fallback)}"
        res = client.send(request)

    return res


def normalize_headers(headers: dict[str, str]) -> dict[str, str]:
    """
    Ensure that known headers are in the correct casing.
    """

    new_headers = {}

    for k, v in headers.items():
        for header_key in const.HeaderKey:
            if k.lower() == header_key.value.lower():
                k = header_key.value

        new_headers[k] = v

    return new_headers


def parse_url(url: str) -> str:
    parsed = urllib.parse.urlparse(url)

    if parsed.scheme == "":
        parsed._replace(scheme="https")

    return parsed.geturl()


async def fetch_with_thready_safety(
    logger: types.Logger,
    client: ThreadAwareAsyncHTTPClient,
    client_sync: httpx.Client,
    request: httpx.Request,
) -> httpx.Response:
    """
    Safely handles the situation where the async HTTP client is called in a
    different thread.
    """

    if client.is_same_thread() is True:
        return await client.send(request)

    # Python freaks out if you call httpx.AsyncClient's async methods in a
    # multiple threads. To solve this, we'll use the synchronous client
    # instead
    logger.warning(
        "called an async client method in a different thread; falling back to synchronous HTTP client"
    )
    return client_sync.send(request)


class RequestSignature:
    _signature: typing.Optional[str] = None
    _timestamp: typing.Optional[int] = None

    def __init__(
        self,
        body: bytes,
        headers: dict[str, str],
        mode: const.ServerKind,
    ) -> None:
        self._body = body
        self._mode = mode

        sig_header = headers.get(const.HeaderKey.SIGNATURE.value)
        if sig_header is not None:
            parsed = urllib.parse.parse_qs(sig_header)
            if "t" in parsed:
                self._timestamp = int(parsed["t"][0])
            if "s" in parsed:
                self._signature = parsed["s"][0]

    def _validate(
        self,
        signing_key: typing.Optional[str],
    ) -> types.MaybeError[None]:
        if self._mode == const.ServerKind.DEV_SERVER:
            return None

        if signing_key is None:
            return errors.SigningKeyMissingError(
                "cannot validate signature in production mode without a signing key"
            )

        if self._signature is None:
            return errors.HeaderMissingError(
                f"cannot validate signature in production mode without a {const.HeaderKey.SIGNATURE.value} header"
            )

        mac = hmac.new(
            transforms.remove_signing_key_prefix(signing_key).encode("utf-8"),
            self._body,
            hashlib.sha256,
        )
        mac.update(str(self._timestamp).encode("utf-8"))
        if not hmac.compare_digest(self._signature, mac.hexdigest()):
            return errors.SigVerificationFailedError()

        return None

    def validate(
        self,
        *,
        signing_key: typing.Optional[str],
        signing_key_fallback: typing.Optional[str],
    ) -> types.MaybeError[None]:
        """
        Validate the request signature. Falls back to the fallback signing key if
        signature validation fails with the primary signing key.

        Args:
        ----
            signing_key: The primary signing key.
            signing_key_fallback: The fallback signing key.
        """

        err = self._validate(signing_key)
        if err is not None and signing_key_fallback is not None:
            # If the signature validation failed but there's a "fallback"
            # signing key, attempt to validate the signature with the fallback
            # key
            err = self._validate(signing_key_fallback)

        return err



"""Tornado integration for Inngest."""

import json
import typing

import tornado.web

from inngest._internal import (
    client_lib,
    comm,
    const,
    errors,
    execution,
    function,
    net,
    transforms,
)

FRAMEWORK = const.Framework.TORNADO


def serve(
    app: tornado.web.Application,
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    serve_origin: typing.Optional[str] = None,
    serve_path: typing.Optional[str] = None,
) -> None:
    """
    Serve Inngest functions in a Tornado app.

    Args:
    ----
        app: Tornado app.
        client: Inngest client.
        functions: List of functions to serve.

        serve_origin: Origin to serve the functions from.
        serve_path: Path to serve the functions from.
    """
    handler = comm.CommHandler(
        api_base_url=client.api_origin,
        client=client,
        framework=FRAMEWORK,
        functions=functions,
    )

    class InngestHandler(tornado.web.RequestHandler):
        def data_received(
            self, chunk: bytes
        ) -> typing.Optional[typing.Awaitable[None]]:
            return None

        def get(self) -> None:
            headers = net.normalize_headers(dict(self.request.headers.items()))

            server_kind = transforms.get_server_kind(headers)
            if isinstance(server_kind, Exception):
                client.logger.error(server_kind)
                server_kind = None

            req_sig = net.RequestSignature(
                body=self.request.body,
                headers=headers,
                mode=client._mode,
            )

            comm_res = handler.inspect(server_kind, req_sig)

            self._write_comm_response(comm_res, server_kind)

        def post(self) -> None:
            fn_id: typing.Optional[str]
            raw_fn_id = self.request.query_arguments.get(
                const.QueryParamKey.FUNCTION_ID.value
            )
            if raw_fn_id is None or len(raw_fn_id) == 0:
                raise errors.QueryParamMissingError(
                    const.QueryParamKey.FUNCTION_ID.value
                )
            fn_id = raw_fn_id[0].decode("utf-8")

            step_id: typing.Optional[str]
            raw_step_id = self.request.query_arguments.get(
                const.QueryParamKey.STEP_ID.value
            )
            if raw_step_id is None or len(raw_step_id) == 0:
                raise errors.QueryParamMissingError(
                    const.QueryParamKey.STEP_ID.value
                )
            step_id = raw_step_id[0].decode("utf-8")

            headers = net.normalize_headers(dict(self.request.headers.items()))

            server_kind = transforms.get_server_kind(headers)
            if isinstance(server_kind, Exception):
                client.logger.error(server_kind)
                server_kind = None

            req_sig = net.RequestSignature(
                body=self.request.body,
                headers=headers,
                mode=client._mode,
            )

            call = execution.Call.from_raw(json.loads(self.request.body))
            if isinstance(call, Exception):
                return self._write_comm_response(
                    comm.CommResponse.from_error(client.logger, call),
                    server_kind,
                )

            comm_res = handler.call_function_sync(
                call=call,
                fn_id=fn_id,
                req_sig=req_sig,
                target_hashed_id=step_id,
            )

            self._write_comm_response(comm_res, server_kind)

        def put(self) -> None:
            headers = net.normalize_headers(dict(self.request.headers.items()))

            server_kind = transforms.get_server_kind(headers)
            if isinstance(server_kind, Exception):
                client.logger.error(server_kind)
                server_kind = None

            sync_id: typing.Optional[str] = None
            raw_sync_id = self.request.query_arguments.get(
                const.QueryParamKey.SYNC_ID.value
            )
            if raw_sync_id is not None:
                sync_id = raw_sync_id[0].decode("utf-8")

            comm_res = handler.register_sync(
                app_url=net.create_serve_url(
                    request_url=self.request.full_url(),
                    serve_origin=serve_origin,
                    serve_path=serve_path,
                ),
                server_kind=server_kind,
                sync_id=sync_id,
            )

            self._write_comm_response(comm_res, server_kind)

        def _write_comm_response(
            self,
            comm_res: comm.CommResponse,
            server_kind: typing.Optional[const.ServerKind],
        ) -> None:
            body = transforms.dump_json(comm_res.body)
            if isinstance(body, Exception):
                comm_res = comm.CommResponse.from_error(client.logger, body)
                body = json.dumps(comm_res.body)

            self.write(body)

            for k, v in comm_res.headers.items():
                self.add_header(k, v)
            for k, v in net.create_headers(
                env=client.env,
                framework=FRAMEWORK,
                server_kind=server_kind,
            ).items():
                self.add_header(k, v)

            self.set_status(comm_res.status_code)

    app.add_handlers(r".*", [("/api/inngest", InngestHandler)])



from __future__ import annotations

import logging
import typing

import pydantic
import typing_extensions

if typing.TYPE_CHECKING:
    # https://github.com/python/typeshed/issues/7855
    Logger = typing.Union[logging.Logger, logging.LoggerAdapter[logging.Logger]]
else:
    Logger = object

T = typing.TypeVar("T")
TTuple = typing_extensions.TypeVarTuple("TTuple")


class EmptySentinel:
    pass


empty_sentinel = EmptySentinel()


# Type checking conditional is necessary because of some mutual exclusivity
# between Mypy and Pydantic. Seems like recursive types are finicky right now,
# but hopefully we can simplify this code as support improves
if typing.TYPE_CHECKING:
    # Mypy uses this statically. Pydantic can't use this at runtime since it'll
    # error with "name 'JSON' is not defined"
    JSON = typing.Union[
        bool,
        float,
        int,
        str,
        typing.Mapping[str, "JSON"],
        typing.Sequence["JSON"],
        None,
    ]
else:
    # Pydantic uses this at runtime. Mypy can't use this since it'll error with
    # "possible cyclic definition"
    JSON = typing_extensions.TypeAliasType(
        "JSON",
        typing.Union[
            bool,
            int,
            float,
            str,
            dict[str, "JSON"],
            list["JSON"],
            None,
        ],
    )


JSONT = typing.TypeVar("JSONT", bound=JSON)


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(strict=True)

    def __init__(
        __pydantic_self__,  # noqa: N805
        *args: object,
        **kwargs: object,
    ) -> None:
        try:
            super().__init__(*args, **kwargs)
        except pydantic.ValidationError as err:
            raise __pydantic_self__.convert_validation_error(err) from None

    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        """
        Override this method in subclasses to convert Pydantic's
        ValidationError into a different error.
        """
        return err

    @classmethod
    def from_raw(
        cls: type[BaseModelT],
        raw: object,
    ) -> typing.Union[BaseModelT, Exception]:
        try:
            return cls.model_validate(raw)
        except Exception as err:
            return err

    def to_dict(self) -> MaybeError[dict[str, object]]:
        try:
            return self.model_dump(by_alias=True, mode="json")
        except Exception as err:
            return err


BaseModelT = typing.TypeVar("BaseModelT", bound=BaseModel)

MaybeError: typing_extensions.TypeAlias = typing.Union[T, Exception]



"""
Even though this file has test functions, they don't actually need to run. This
file is purely for static type analysis
"""

import typing

import inngest

_T = typing.TypeVar("_T")


class _AssertType(typing.Generic[_T]):
    """
    Used to assert type hints. A no-op at runtime.
    """

    def __init__(self, value: _T) -> None:
        pass


client = inngest.Inngest(app_id="foo", is_production=False)


def sync_fn_with_sync_step() -> None:
    """
    Test that a sync function cannot use an async step type
    """

    @client.create_function(  # type: ignore[arg-type]
        fn_id="foo",
        trigger=inngest.TriggerEvent(event="foo"),
    )
    def fn(
        ctx: inngest.Context,
        step: inngest.Step,
    ) -> None:
        pass


def async_fn_with_sync_step() -> None:
    """
    Test that an async function cannot use a sync step type
    """

    @client.create_function(  # type: ignore[arg-type]
        fn_id="foo",
        trigger=inngest.TriggerEvent(event="foo"),
    )
    async def fn(
        ctx: inngest.Context,
        step: inngest.StepSync,
    ) -> None:
        pass


def event_data_dict() -> None:
    """
    Test that event data can be a dict. This is exists to prevent a regression
    after a type fix where `Event.data` was typed as `dict[str, object]`, which
    precluded more specific types like `dict[str, str]`. That happens because
    `dict` is considered to be mutable, so type checkers make it invariant
    """

    @client.create_function(
        fn_id="foo",
        trigger=inngest.TriggerEvent(event="foo"),
    )
    async def fn(
        ctx: inngest.Context,
        step: inngest.Step,
    ) -> None:
        data = {"foo": "bar"}

        await step.send_event(
            "foo",
            inngest.Event(data=data, name="foo"),
        )


def step_callback_Args() -> None:
    """
    Ensure that step callback arguments are correctly typed. They're inferred
    from the step callback's type signature.
    """

    def step_callback_async(a: int, b: str) -> bool:
        return True

    def step_callback_sync(a: int, b: str) -> bool:
        return True

    @client.create_function(
        fn_id="foo",
        trigger=inngest.TriggerEvent(event="foo"),
    )
    async def async_fn(
        ctx: inngest.Context,
        step: inngest.Step,
    ) -> None:
        # Incorrect arg types fail type checker
        await step.run(
            "step",
            step_callback_async,  # type: ignore[arg-type]
            "a",
            1,
        )  # type: ignore[call-arg]

        output = await step.run("step", step_callback_async, 1, "a")

        # Output type is correctly inferred
        _AssertType[bool](output)
        _AssertType[str](output)  # type: ignore[arg-type]

    @client.create_function(
        fn_id="foo",
        trigger=inngest.TriggerEvent(event="foo"),
    )
    async def async_fn_with_sync_callback(
        ctx: inngest.Context,
        step: inngest.Step,
    ) -> None:
        # Incorrect arg types fail type checker
        await step.run(
            "step",
            step_callback_sync,  # type: ignore[arg-type]
            "a",
            1,
        )  # type: ignore[call-arg]

        output = await step.run("step", step_callback_sync, 1, "a")

        # Output type is correctly inferred
        _AssertType[bool](output)
        _AssertType[str](output)  # type: ignore[arg-type]

    @client.create_function(
        fn_id="foo",
        trigger=inngest.TriggerEvent(event="foo"),
    )
    def sync_fn(
        ctx: inngest.Context,
        step: inngest.StepSync,
    ) -> None:
        # Incorrect arg types fail type checker
        step.run(
            "step",
            step_callback_sync,  # type: ignore[arg-type]
            "a",
            1,
        )

        output = step.run("step", step_callback_sync, 1, "a")

        # Output type is correctly inferred
        _AssertType[bool](output)
        _AssertType[str](output)  # type: ignore[arg-type]



from .manager import MiddlewareManager
from .middleware import Middleware, MiddlewareSync, UninitializedMiddleware

__all__ = [
    "Middleware",
    "MiddlewareManager",
    "MiddlewareSync",
    "UninitializedMiddleware",
]



from __future__ import annotations

import dataclasses
import inspect
import typing
import urllib.parse

import pydantic
import typing_extensions

from inngest._internal import (
    client_lib,
    const,
    errors,
    event_lib,
    execution,
    function_config,
    middleware_lib,
    step_lib,
    transforms,
    types,
)


@dataclasses.dataclass
class Context:
    # TODO: Remove this in v0.4.0. It's only here to avoid a breaking change to
    # Middleware.transform_input
    _steps: step_lib.StepMemos

    attempt: int
    event: event_lib.Event
    events: list[event_lib.Event]
    logger: types.Logger
    run_id: str


@dataclasses.dataclass
class _Config:
    # The user-defined function
    main: function_config.FunctionConfig

    # The internal on_failure function
    on_failure: typing.Optional[function_config.FunctionConfig]


@typing.runtime_checkable
class FunctionHandlerAsync(typing.Protocol):
    def __call__(
        self,
        ctx: Context,
        step: step_lib.Step,
    ) -> typing.Awaitable[types.JSON]:
        ...


@typing.runtime_checkable
class FunctionHandlerSync(typing.Protocol):
    def __call__(
        self,
        ctx: Context,
        step: step_lib.StepSync,
    ) -> types.JSON:
        ...


def _is_function_handler_async(
    value: typing.Union[FunctionHandlerAsync, FunctionHandlerSync],
) -> typing_extensions.TypeGuard[FunctionHandlerAsync]:
    return inspect.iscoroutinefunction(value)


def _is_function_handler_sync(
    value: typing.Union[FunctionHandlerAsync, FunctionHandlerSync],
) -> typing_extensions.TypeGuard[FunctionHandlerSync]:
    return not inspect.iscoroutinefunction(value)


class FunctionOpts(types.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    batch_events: typing.Optional[function_config.Batch] = None
    cancel: typing.Optional[list[function_config.Cancel]] = None
    concurrency: typing.Optional[list[function_config.Concurrency]] = None
    debounce: typing.Optional[function_config.Debounce] = None

    # Unique within an environment
    fully_qualified_id: str

    # Unique within an app
    local_id: str

    name: str
    on_failure: typing.Union[
        FunctionHandlerAsync, FunctionHandlerSync, None
    ] = None
    rate_limit: typing.Optional[function_config.RateLimit] = None
    retries: typing.Optional[int] = None
    throttle: typing.Optional[function_config.Throttle] = None

    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        return errors.FunctionConfigInvalidError.from_validation_error(err)


class Function:
    _handler: typing.Union[FunctionHandlerAsync, FunctionHandlerSync]
    _on_failure_fn_id: typing.Optional[str] = None
    _opts: FunctionOpts
    _triggers: list[
        typing.Union[function_config.TriggerCron, function_config.TriggerEvent]
    ]

    @property
    def id(self) -> str:
        return self._opts.fully_qualified_id

    @property
    def is_handler_async(self) -> bool:
        """Whether the main handler is async."""
        return _is_function_handler_async(self._handler)

    @property
    def is_on_failure_handler_async(self) -> typing.Optional[bool]:
        """
        Whether the on_failure handler is async. Returns None if there isn't an
        on_failure handler.
        """
        if self._opts.on_failure is None:
            return None
        return _is_function_handler_async(self._opts.on_failure)

    @property
    def on_failure_fn_id(self) -> typing.Optional[str]:
        return self._on_failure_fn_id

    def __init__(
        self,
        opts: FunctionOpts,
        trigger: typing.Union[
            function_config.TriggerCron,
            function_config.TriggerEvent,
            list[
                typing.Union[
                    function_config.TriggerCron, function_config.TriggerEvent
                ]
            ],
        ],
        handler: typing.Union[FunctionHandlerAsync, FunctionHandlerSync],
        middleware: typing.Optional[
            list[middleware_lib.UninitializedMiddleware]
        ] = None,
    ) -> None:
        self._handler = handler
        self._middleware = middleware or []
        self._opts = opts
        self._triggers = trigger if isinstance(trigger, list) else [trigger]

        if opts.on_failure is not None:
            self._on_failure_fn_id = f"{opts.fully_qualified_id}-failure"

    async def call(
        self,
        client: client_lib.Inngest,
        ctx: Context,
        fn_id: str,
        middleware: middleware_lib.MiddlewareManager,
        target_hashed_id: typing.Optional[str],
    ) -> execution.CallResult:
        middleware = middleware_lib.MiddlewareManager.from_manager(middleware)
        for m in self._middleware:
            middleware.add(m)

        # Give middleware the opportunity to change some of params passed to the
        # user's handler.
        new_ctx = await middleware.transform_input(ctx)
        if isinstance(new_ctx, Exception):
            return execution.CallError.from_error(new_ctx)
        ctx = new_ctx

        # No memoized data means we're calling the function for the first time.
        if ctx._steps.size == 0:
            err = await middleware.before_execution()
            if isinstance(err, Exception):
                return execution.CallError.from_error(err)

        try:
            handler: typing.Union[FunctionHandlerAsync, FunctionHandlerSync]
            if self.id == fn_id:
                handler = self._handler
            elif self.on_failure_fn_id == fn_id:
                if self._opts.on_failure is None:
                    return execution.CallError.from_error(
                        errors.FunctionNotFoundError("on_failure not defined")
                    )
                handler = self._opts.on_failure
            else:
                return execution.CallError.from_error(
                    errors.FunctionNotFoundError("function ID mismatch")
                )

            output: object

            try:
                # # Determine whether the handler is async (i.e. if we need to await
                # # it). Sync functions are OK in async contexts, so it's OK if the
                # # handler is sync.
                if _is_function_handler_async(handler):
                    output = await handler(
                        ctx=ctx,
                        step=step_lib.Step(
                            client,
                            ctx._steps,
                            middleware,
                            step_lib.StepIDCounter(),
                            target_hashed_id,
                        ),
                    )
                elif _is_function_handler_sync(handler):
                    output = handler(
                        ctx=ctx,
                        step=step_lib.StepSync(
                            client,
                            ctx._steps,
                            middleware,
                            step_lib.StepIDCounter(),
                            target_hashed_id,
                        ),
                    )
                else:
                    # Should be unreachable but Python's custom type guards don't
                    # support negative checks :(
                    return execution.CallError.from_error(
                        errors.UnknownError(
                            "unable to determine function handler type"
                        )
                    )
            except Exception as user_err:
                transforms.remove_first_traceback_frame(user_err)
                raise _UserError(user_err)

            err = await middleware.after_execution()
            if isinstance(err, Exception):
                return execution.CallError.from_error(err)

            output = await middleware.transform_output(
                # Function output isn't wrapped in an Output object, so we need
                # to wrap it to make it compatible with middleware.
                execution.Output(data=output)
            )
            if isinstance(output, Exception):
                return execution.CallError.from_error(output)

            if output is None:
                return execution.FunctionCallResponse(data=None)
            return execution.FunctionCallResponse(data=output.data)
        except step_lib.ResponseInterrupt as interrupt:
            err = await middleware.after_execution()
            if isinstance(err, Exception):
                return execution.CallError.from_error(err)

            # TODO: How should transform_output work with multiple responses?
            if len(interrupt.responses) == 1:
                output = await middleware.transform_output(
                    interrupt.responses[0].data
                )
                if isinstance(output, Exception):
                    return execution.CallError.from_error(output)
                interrupt.responses[0].data = output

            return interrupt.responses
        except _UserError as err:
            return execution.CallError.from_error(err.err)
        except step_lib.SkipInterrupt as err:
            # This should only happen in a non-deterministic scenario, where
            # step targeting is enabled and an unexpected step is encountered.
            # We don't currently have a way to recover from this scenario.

            return execution.CallError.from_error(
                errors.StepUnexpectedError(
                    f'found step "{err.step_id}" when targeting a different step'
                )
            )
        except Exception as err:
            return execution.CallError.from_error(err)

    def call_sync(
        self,
        client: client_lib.Inngest,
        ctx: Context,
        fn_id: str,
        middleware: middleware_lib.MiddlewareManager,
        target_hashed_id: typing.Optional[str],
    ) -> execution.CallResult:
        middleware = middleware_lib.MiddlewareManager.from_manager(middleware)
        for m in self._middleware:
            middleware.add(m)

        # Give middleware the opportunity to change some of params passed to the
        # user's handler.
        new_ctx = middleware.transform_input_sync(ctx)
        if isinstance(new_ctx, Exception):
            return execution.CallError.from_error(new_ctx)
        ctx = new_ctx

        # No memoized data means we're calling the function for the first time.
        if ctx._steps.size == 0:
            middleware.before_execution_sync()

        try:
            handler: typing.Union[FunctionHandlerAsync, FunctionHandlerSync]
            if self.id == fn_id:
                handler = self._handler
            elif self.on_failure_fn_id == fn_id:
                if self._opts.on_failure is None:
                    return execution.CallError.from_error(
                        errors.FunctionNotFoundError("on_failure not defined")
                    )
                handler = self._opts.on_failure
            else:
                return execution.CallError.from_error(
                    errors.FunctionNotFoundError("function ID mismatch")
                )

            if _is_function_handler_sync(handler):
                try:
                    output: object = handler(
                        ctx=ctx,
                        step=step_lib.StepSync(
                            client,
                            ctx._steps,
                            middleware,
                            step_lib.StepIDCounter(),
                            target_hashed_id,
                        ),
                    )
                except Exception as user_err:
                    transforms.remove_first_traceback_frame(user_err)
                    raise _UserError(user_err)
            else:
                return execution.CallError.from_error(
                    errors.AsyncUnsupportedError(
                        "encountered async function in non-async context"
                    )
                )

            err = middleware.after_execution_sync()
            if isinstance(err, Exception):
                return execution.CallError.from_error(err)

            output = middleware.transform_output_sync(
                # Function output isn't wrapped in an Output object, so we need
                # to wrap it to make it compatible with middleware.
                execution.Output(data=output)
            )
            if isinstance(output, Exception):
                return execution.CallError.from_error(output)

            if output is None:
                return execution.FunctionCallResponse(data=None)
            return execution.FunctionCallResponse(data=output.data)
        except step_lib.ResponseInterrupt as interrupt:
            err = middleware.after_execution_sync()
            if isinstance(err, Exception):
                return execution.CallError.from_error(err)

            # TODO: How should transform_output work with multiple responses?
            if len(interrupt.responses) == 1:
                output = middleware.transform_output_sync(
                    interrupt.responses[0].data
                )
                if isinstance(output, Exception):
                    return execution.CallError.from_error(output)
                interrupt.responses[0].data = output

            return interrupt.responses
        except _UserError as err:
            return execution.CallError.from_error(err.err)
        except step_lib.SkipInterrupt as err:
            # This should only happen in a non-deterministic scenario, where
            # step targeting is enabled and an unexpected step is encountered.
            # We don't currently have a way to recover from this scenario.

            return execution.CallError.from_error(
                errors.StepUnexpectedError(
                    f'found step "{err.step_id}" when targeting a different step'
                )
            )
        except Exception as err:
            return execution.CallError.from_error(err)

    def get_config(self, app_url: str) -> _Config:
        fn_id = self._opts.fully_qualified_id
        name = self._opts.name

        if self._opts.retries is not None:
            retries = function_config.Retries(attempts=self._opts.retries)
        else:
            retries = None

        url = (
            app_url
            + "?"
            + urllib.parse.urlencode(
                {
                    "fnId": fn_id,
                    "stepId": const.ROOT_STEP_ID,
                }
            )
        )

        main = function_config.FunctionConfig(
            batch_events=self._opts.batch_events,
            cancel=self._opts.cancel,
            concurrency=self._opts.concurrency,
            debounce=self._opts.debounce,
            id=fn_id,
            name=name,
            rate_limit=self._opts.rate_limit,
            steps={
                const.ROOT_STEP_ID: function_config.Step(
                    id=const.ROOT_STEP_ID,
                    name=const.ROOT_STEP_ID,
                    retries=retries,
                    runtime=function_config.Runtime(
                        type="http",
                        url=url,
                    ),
                ),
            },
            throttle=self._opts.throttle,
            triggers=self._triggers,
        )

        on_failure = None
        if self.on_failure_fn_id is not None:
            url = (
                app_url
                + "?"
                + urllib.parse.urlencode(
                    {
                        "fnId": self.on_failure_fn_id,
                        "stepId": const.ROOT_STEP_ID,
                    }
                )
            )

            on_failure = function_config.FunctionConfig(
                id=self.on_failure_fn_id,
                name=f"{name} (failure)",
                steps={
                    const.ROOT_STEP_ID: function_config.Step(
                        id=const.ROOT_STEP_ID,
                        name=const.ROOT_STEP_ID,
                        retries=function_config.Retries(attempts=0),
                        runtime=function_config.Runtime(
                            type="http",
                            url=url,
                        ),
                    )
                },
                triggers=[
                    function_config.TriggerEvent(
                        event=const.InternalEvents.FUNCTION_FAILED.value,
                        expression=f"event.data.function_id == '{self.id}'",
                    )
                ],
            )

        return _Config(main=main, on_failure=on_failure)

    def get_id(self) -> str:
        return self._opts.fully_qualified_id


class _UserError(Exception):
    """
    Wrap an error that occurred in user code.
    """

    def __init__(self, err: Exception) -> None:
        self.err = err



"""FastAPI integration for Inngest."""

import json
import typing

import fastapi

from ._internal import (
    client_lib,
    comm,
    const,
    execution,
    function,
    net,
    transforms,
)

FRAMEWORK = const.Framework.FAST_API


def serve(
    app: fastapi.FastAPI,
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    serve_origin: typing.Optional[str] = None,
    serve_path: typing.Optional[str] = None,
) -> None:
    """
    Serve Inngest functions in a FastAPI app.

    Args:
    ----
        app: FastAPI app.
        client: Inngest client.
        functions: List of functions to serve.

        serve_origin: Origin to serve the functions from.
        serve_path: Path to serve the functions from.
    """

    handler = comm.CommHandler(
        api_base_url=client.api_origin,
        client=client,
        framework=FRAMEWORK,
        functions=functions,
    )

    @app.get("/api/inngest")
    async def get_api_inngest(
        request: fastapi.Request,
    ) -> fastapi.Response:
        body = await request.body()
        headers = net.normalize_headers(dict(request.headers.items()))

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            client.logger.error(server_kind)
            server_kind = None

        return _to_response(
            client,
            handler.inspect(
                server_kind,
                net.RequestSignature(
                    body=body,
                    headers=headers,
                    mode=client._mode,
                ),
            ),
            server_kind,
        )

    @app.post("/api/inngest")
    async def post_inngest_api(
        fnId: str,  # noqa: N803
        stepId: str,  # noqa: N803
        request: fastapi.Request,
    ) -> fastapi.Response:
        body = await request.body()
        headers = net.normalize_headers(dict(request.headers.items()))

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            client.logger.error(server_kind)
            server_kind = None

        call = execution.Call.from_raw(json.loads(body))
        if isinstance(call, Exception):
            return _to_response(
                client,
                comm.CommResponse.from_error(client.logger, call),
                server_kind,
            )

        return _to_response(
            client,
            await handler.call_function(
                call=call,
                fn_id=fnId,
                req_sig=net.RequestSignature(
                    body=body,
                    headers=headers,
                    mode=client._mode,
                ),
                target_hashed_id=stepId,
            ),
            server_kind,
        )

    @app.put("/api/inngest")
    async def put_inngest_api(
        request: fastapi.Request,
    ) -> fastapi.Response:
        headers = net.normalize_headers(dict(request.headers.items()))

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            client.logger.error(server_kind)
            server_kind = None

        sync_id = request.query_params.get(const.QueryParamKey.SYNC_ID.value)

        return _to_response(
            client,
            await handler.register(
                app_url=net.create_serve_url(
                    request_url=str(request.url),
                    serve_origin=serve_origin,
                    serve_path=serve_path,
                ),
                server_kind=server_kind,
                sync_id=sync_id,
            ),
            server_kind,
        )


def _to_response(
    client: client_lib.Inngest,
    comm_res: comm.CommResponse,
    server_kind: typing.Union[const.ServerKind, None],
) -> fastapi.responses.Response:
    body = transforms.dump_json(comm_res.body)
    if isinstance(body, Exception):
        comm_res = comm.CommResponse.from_error(client.logger, body)
        body = json.dumps(comm_res.body)

    return fastapi.responses.Response(
        content=body.encode("utf-8"),
        headers={
            **comm_res.headers,
            **net.create_headers(
                env=client.env,
                framework=FRAMEWORK,
                server_kind=server_kind,
            ),
        },
        status_code=comm_res.status_code,
    )



"""Public entrypoint for the Inngest SDK."""


from ._internal.client_lib import Inngest
from ._internal.errors import NonRetriableError, StepError
from ._internal.event_lib import Event
from ._internal.execution import Output
from ._internal.function import Context, Function
from ._internal.function_config import (
    Batch,
    Cancel,
    Concurrency,
    Debounce,
    RateLimit,
    Throttle,
    TriggerCron,
    TriggerEvent,
)
from ._internal.middleware_lib import Middleware, MiddlewareSync
from ._internal.step_lib import FunctionID, Step, StepSync
from ._internal.types import JSON

__all__ = [
    "Batch",
    "Cancel",
    "Concurrency",
    "Context",
    "Debounce",
    "Event",
    "Function",
    "FunctionID",
    "Inngest",
    "JSON",
    "Middleware",
    "MiddlewareSync",
    "NonRetriableError",
    "Output",
    "RateLimit",
    "Step",
    "StepError",
    "StepSync",
    "Throttle",
    "TriggerCron",
    "TriggerEvent",
]



"""
Encryption middleware for Inngest.
"""

from __future__ import annotations

import json
import typing

import nacl.encoding
import nacl.secret
import nacl.utils

import inngest

encryption_marker: typing.Final = "__ENCRYPTED__"


class EncryptionMiddleware(inngest.MiddlewareSync):
    """
    Middleware that encrypts and decrypts data using a symmetric key. The
    cryptography library is a dependency.
    """

    def __init__(
        self,
        client: inngest.Inngest,
        secret_key: typing.Union[bytes, str],
    ) -> None:
        """
        Args:
        ----
            client: Inngest client.
            secret_key: Fernet secret key used for encryption and decryption.
        """

        super().__init__(client)

        if isinstance(secret_key, str):
            secret_key = bytes.fromhex(secret_key)

        self._box = nacl.secret.SecretBox(secret_key)

    @classmethod
    def factory(
        cls,
        secret_key: typing.Union[bytes, str],
    ) -> typing.Callable[[inngest.Inngest], EncryptionMiddleware]:
        """
        Create an encryption middleware factory that can be passed to an Inngest
        client or function.

        Args:
        ----
            secret_key: Fernet secret key used for encryption and decryption.
        """

        def _factory(client: inngest.Inngest) -> EncryptionMiddleware:
            return cls(client, secret_key)

        return _factory

    def _encrypt(self, data: object) -> dict[str, typing.Union[bool, str]]:
        if isinstance(data, dict) and data.get(encryption_marker) is True:
            # Already encrypted
            return data

        byt = json.dumps(data).encode()
        ciphertext = self._box.encrypt(
            byt,
            encoder=nacl.encoding.Base64Encoder,
        )
        return {
            encryption_marker: True,
            "data": ciphertext.decode(),
        }

    def _decrypt(self, data: object) -> inngest.JSON:
        if not _is_encrypted(data) or not isinstance(data, dict):
            # Not encrypted
            return data  # type: ignore

        encrypted = data.get("data")
        if not isinstance(encrypted, str):
            return data

        byt = self._box.decrypt(
            encrypted.encode(),
            encoder=nacl.encoding.Base64Encoder,
        )
        return json.loads(byt.decode())  # type: ignore

    def _decrypt_event_data(
        self,
        data: typing.Mapping[str, inngest.JSON],
    ) -> typing.Mapping[str, inngest.JSON]:
        is_everything_encrypted = _is_encrypted(data)
        if is_everything_encrypted:
            decrypted = self._decrypt(data)
            if isinstance(decrypted, dict):
                data = decrypted
        else:
            data = {k: self._decrypt(v) for k, v in data.items()}

        return data

    def before_send_events(self, events: list[inngest.Event]) -> None:
        """
        Encrypt event data before sending it to the Inngest server.
        """

        for event in events:
            event.data = self._encrypt(event.data)

    def transform_input(self, ctx: inngest.Context) -> inngest.Context:
        """
        Decrypt data from the Inngest server.
        """

        for step in ctx._steps.values():
            step.data = self._decrypt(step.data)

        ctx.event.data = self._decrypt_event_data(ctx.event.data)

        for event in ctx.events:
            event.data = self._decrypt_event_data(event.data)

        return ctx

    def transform_output(self, output: inngest.Output) -> inngest.Output:
        """
        Encrypt data before sending it to the Inngest server.
        """

        output.data = self._encrypt(output.data)
        return output


def _is_encrypted(value: object) -> bool:
    return isinstance(value, dict) and value.get(encryption_marker) is True



import enum

import pydantic

from . import const, function_config, types


class DeployType(enum.Enum):
    PING = "ping"


class RegisterRequest(types.BaseModel):
    app_name: str = pydantic.Field(..., serialization_alias="appname")
    deploy_type: DeployType
    framework: const.Framework
    functions: list[function_config.FunctionConfig] = pydantic.Field(
        min_length=1
    )
    sdk: str
    url: str
    v: str



from __future__ import annotations

import enum
import typing

import pydantic
import typing_extensions

from . import const, errors, event_lib, transforms, types


class Call(types.BaseModel):
    ctx: CallContext
    event: event_lib.Event
    events: typing.Optional[list[event_lib.Event]] = None
    steps: dict[str, object]
    use_api: bool


class CallContext(types.BaseModel):
    attempt: int
    run_id: str
    stack: CallStack


class CallStack(types.BaseModel):
    stack: list[str]


class CallError(types.BaseModel):
    """
    When an error that occurred during a call. Used for both function- and step-level
    errors.
    """

    code: const.ErrorCode
    is_retriable: bool
    message: str
    name: str
    original_error: object = pydantic.Field(exclude=True)
    stack: typing.Optional[str]
    step_id: typing.Optional[str]

    @classmethod
    def from_error(
        cls,
        err: Exception,
        step_id: typing.Optional[str] = None,
    ) -> CallError:
        code = const.ErrorCode.UNKNOWN
        if isinstance(err, errors.Error):
            code = err.code
            is_retriable = err.is_retriable
            message = err.message
            name = err.name
            stack = err.stack
        else:
            is_retriable = True
            message = str(err)
            name = type(err).__name__
            stack = transforms.get_traceback(err)

        return cls(
            code=code,
            is_retriable=is_retriable,
            message=message,
            name=name,
            original_error=err,
            stack=stack,
            step_id=step_id,
        )


class FunctionCallResponse(types.BaseModel):
    """When a function successfully returns."""

    data: object


class StepResponse(types.BaseModel):
    data: typing.Optional[Output] = None
    display_name: str = pydantic.Field(..., serialization_alias="displayName")
    id: str

    # Deprecated
    name: typing.Optional[str] = None

    op: Opcode
    opts: typing.Optional[dict[str, object]] = None


class MemoizedError(types.BaseModel):
    message: str
    name: str
    stack: typing.Optional[str] = None

    @classmethod
    def from_error(cls, err: Exception) -> MemoizedError:
        return cls(
            message=str(err),
            name=type(err).__name__,
            stack=transforms.get_traceback(err),
        )


class Output(types.BaseModel):
    # Fail validation if any extra fields exist, because this will prevent
    # accidentally assuming user data is nested data
    model_config = pydantic.ConfigDict(extra="forbid")

    data: object = None

    # TODO: Change the type to MemoizedError. But that requires a breaking
    # change, so do it in version 0.4
    error: typing.Optional[dict[str, object]] = None


def is_step_call_responses(
    value: object,
) -> typing_extensions.TypeGuard[list[StepResponse]]:
    if not isinstance(value, list):
        return False
    return all(isinstance(item, StepResponse) for item in value)


CallResult: typing_extensions.TypeAlias = typing.Union[
    list[StepResponse], FunctionCallResponse, CallError
]


class Opcode(enum.Enum):
    INVOKE = "InvokeFunction"
    PLANNED = "StepPlanned"
    SLEEP = "Sleep"
    STEP_RUN = "StepRun"
    STEP_ERROR = "StepError"
    WAIT_FOR_EVENT = "WaitForEvent"


# If the Executor sends this step ID then it isn't targeting a specific step.
UNSPECIFIED_STEP_ID = "step"



import enum
import typing

DEFAULT_API_ORIGIN: typing.Final = "https://api.inngest.com/"
DEFAULT_EVENT_ORIGIN: typing.Final = "https://inn.gs/"
DEV_SERVER_ORIGIN: typing.Final = "http://127.0.0.1:8288/"
LANGUAGE: typing.Final = "py"
ROOT_STEP_ID: typing.Final = "step"
VERSION: typing.Final = "0.3.13.dev2"


class EnvKey(enum.Enum):
    API_BASE_URL = "INNGEST_API_BASE_URL"
    DEV = "INNGEST_DEV"
    EVENT_API_BASE_URL = "INNGEST_EVENT_API_BASE_URL"
    EVENT_KEY = "INNGEST_EVENT_KEY"
    ENV = "INNGEST_ENV"

    # Railway deployment's git branch
    # https://docs.railway.app/develop/variables#railway-provided-variables
    RAILWAY_GIT_BRANCH = "RAILWAY_GIT_BRANCH"

    # Render deployment's git branch
    # https://render.com/docs/environment-variables#all-services
    RENDER_GIT_BRANCH = "RENDER_GIT_BRANCH"

    SERVE_ORIGIN = "INNGEST_SERVE_ORIGIN"
    SERVE_PATH = "INNGEST_SERVE_PATH"
    SIGNING_KEY = "INNGEST_SIGNING_KEY"
    SIGNING_KEY_FALLBACK = "INNGEST_SIGNING_KEY_FALLBACK"

    # Vercel deployment's git branch
    # https://vercel.com/docs/concepts/projects/environment-variables/system-environment-variables#system-environment-variables
    VERCEL_GIT_BRANCH = "VERCEL_GIT_COMMIT_REF"


class ErrorCode(enum.Enum):
    ASYNC_UNSUPPORTED = "async_unsupported"
    BODY_INVALID = "body_invalid"
    EVENT_KEY_UNSPECIFIED = "event_key_unspecified"
    FUNCTION_CONFIG_INVALID = "function_config_invalid"
    FUNCTION_NOT_FOUND = "function_not_found"
    HEADER_MISSING = "header_missing"
    NON_RETRIABLE_ERROR = "non_retriable_error"
    OUTPUT_UNSERIALIZABLE = "output_unserializable"
    QUERY_PARAM_MISSING = "query_param_missing"
    REGISTRATION_FAILED = "registration_failed"
    SERVER_KIND_MISMATCH = "server_kind_mismatch"
    SIGNING_KEY_UNSPECIFIED = "signing_key_unspecified"
    SIG_VERIFICATION_FAILED = "sig_verification_failed"
    STEP_ERRORED = "step_errored"
    STEP_UNEXPECTED = "step_unexpected"
    UNKNOWN = "unknown"
    URL_INVALID = "url_invalid"


class Framework(enum.Enum):
    DIGITAL_OCEAN = "digitalocean"
    DJANGO = "django"
    FAST_API = "fast_api"
    FLASK = "flask"
    TORNADO = "tornado"


class HeaderKey(enum.Enum):
    AUTHORIZATION = "Authorization"
    CONTENT_TYPE = "Content-Type"
    ENV = "X-Inngest-Env"
    EXPECTED_SERVER_KIND = "X-Inngest-Expected-Server-Kind"
    FRAMEWORK = "X-Inngest-Framework"
    NO_RETRY = "X-Inngest-No-Retry"
    SDK = "X-Inngest-SDK"
    SERVER_KIND = "X-Inngest-Server-Kind"
    SERVER_TIMING = "Server-Timing"
    SIGNATURE = "X-Inngest-Signature"
    USER_AGENT = "User-Agent"


class QueryParamKey(enum.Enum):
    FUNCTION_ID = "fnId"
    STEP_ID = "stepId"
    SYNC_ID = "deployId"


class InternalEvents(enum.Enum):
    FUNCTION_FAILED = "inngest/function.failed"


class ServerKind(enum.Enum):
    CLOUD = "cloud"
    DEV_SERVER = "dev"



"""
Django integration for Inngest.
"""

import http
import json
import typing

import django
import django.http
import django.urls
import django.views.decorators.csrf

from ._internal import (
    client_lib,
    comm,
    const,
    errors,
    execution,
    function,
    net,
    transforms,
)

FRAMEWORK = const.Framework.DJANGO


def serve(
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    async_mode: bool = False,
    serve_origin: typing.Optional[str] = None,
    serve_path: typing.Optional[str] = None,
) -> django.urls.URLPattern:
    """
    Serve Inngest functions in a Django app.

    Args:
    ----
        client: Inngest client.
        functions: List of functions to serve.

        async_mode: Whether to serve functions asynchronously.
        serve_origin: Origin to serve Inngest from.
        serve_path: Path to serve Inngest from.
    """

    handler = comm.CommHandler(
        api_base_url=client.api_origin,
        client=client,
        framework=FRAMEWORK,
        functions=functions,
    )

    if async_mode:
        return _create_handler_async(
            client,
            handler,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )
    else:
        return _create_handler_sync(
            client,
            handler,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )


def _create_handler_sync(
    client: client_lib.Inngest,
    handler: comm.CommHandler,
    *,
    serve_origin: typing.Optional[str],
    serve_path: typing.Optional[str],
) -> django.urls.URLPattern:
    def inngest_api(
        request: django.http.HttpRequest,
    ) -> django.http.HttpResponse:
        headers = net.normalize_headers(dict(request.headers.items()))

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            client.logger.error(server_kind)
            server_kind = None

        req_sig = net.RequestSignature(
            body=request.body,
            headers=headers,
            mode=client._mode,
        )

        if request.method == "GET":
            return _to_response(
                client,
                handler.inspect(server_kind, req_sig),
                server_kind,
            )

        if request.method == "POST":
            fn_id = request.GET.get(const.QueryParamKey.FUNCTION_ID.value)
            if fn_id is None:
                raise errors.QueryParamMissingError(
                    const.QueryParamKey.FUNCTION_ID.value
                )

            step_id = request.GET.get(const.QueryParamKey.STEP_ID.value)
            if step_id is None:
                raise errors.QueryParamMissingError(
                    const.QueryParamKey.STEP_ID.value
                )

            call = execution.Call.from_raw(json.loads(request.body))
            if isinstance(call, Exception):
                return _to_response(
                    client,
                    comm.CommResponse.from_error(client.logger, call),
                    server_kind,
                )

            return _to_response(
                client,
                handler.call_function_sync(
                    call=call,
                    fn_id=fn_id,
                    req_sig=req_sig,
                    target_hashed_id=step_id,
                ),
                server_kind,
            )

        if request.method == "PUT":
            sync_id = request.GET.get(const.QueryParamKey.SYNC_ID.value)

            return _to_response(
                client,
                handler.register_sync(
                    app_url=net.create_serve_url(
                        request_url=request.build_absolute_uri(),
                        serve_origin=serve_origin,
                        serve_path=serve_path,
                    ),
                    server_kind=server_kind,
                    sync_id=sync_id,
                ),
                server_kind,
            )

        return django.http.JsonResponse(
            {"error": "Unsupported method"},
            status=http.HTTPStatus.METHOD_NOT_ALLOWED,
        )

    return django.urls.path(
        "api/inngest",
        django.views.decorators.csrf.csrf_exempt(inngest_api),
    )


def _create_handler_async(
    client: client_lib.Inngest,
    handler: comm.CommHandler,
    *,
    serve_origin: typing.Optional[str],
    serve_path: typing.Optional[str],
) -> django.urls.URLPattern:
    async def inngest_api(
        request: django.http.HttpRequest,
    ) -> django.http.HttpResponse:
        headers = net.normalize_headers(dict(request.headers.items()))

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            client.logger.error(server_kind)
            server_kind = None

        req_sig = net.RequestSignature(
            body=request.body,
            headers=headers,
            mode=client._mode,
        )

        if request.method == "GET":
            return _to_response(
                client,
                handler.inspect(server_kind, req_sig),
                server_kind,
            )

        if request.method == "POST":
            fn_id = request.GET.get(const.QueryParamKey.FUNCTION_ID.value)
            if fn_id is None:
                raise errors.QueryParamMissingError(
                    const.QueryParamKey.FUNCTION_ID.value
                )

            step_id = request.GET.get(const.QueryParamKey.STEP_ID.value)
            if step_id is None:
                raise errors.QueryParamMissingError(
                    const.QueryParamKey.STEP_ID.value
                )

            call = execution.Call.from_raw(json.loads(request.body))
            if isinstance(call, Exception):
                return _to_response(
                    client,
                    comm.CommResponse.from_error(client.logger, call),
                    server_kind,
                )

            return _to_response(
                client,
                await handler.call_function(
                    call=call,
                    fn_id=fn_id,
                    req_sig=req_sig,
                    target_hashed_id=step_id,
                ),
                server_kind,
            )

        if request.method == "PUT":
            sync_id = request.GET.get(const.QueryParamKey.SYNC_ID.value)

            return _to_response(
                client,
                await handler.register(
                    app_url=net.create_serve_url(
                        request_url=request.build_absolute_uri(),
                        serve_origin=serve_origin,
                        serve_path=serve_path,
                    ),
                    server_kind=server_kind,
                    sync_id=sync_id,
                ),
                server_kind,
            )

        return django.http.JsonResponse(
            {"error": "Unsupported method"},
            status=http.HTTPStatus.METHOD_NOT_ALLOWED,
        )

    return django.urls.path(
        "api/inngest",
        django.views.decorators.csrf.csrf_exempt(inngest_api),
    )


def _to_response(
    client: client_lib.Inngest,
    comm_res: comm.CommResponse,
    server_kind: typing.Optional[const.ServerKind],
) -> django.http.HttpResponse:
    body = transforms.dump_json(comm_res.body)
    if isinstance(body, Exception):
        comm_res = comm.CommResponse.from_error(client.logger, body)
        body = json.dumps(comm_res.body)

    return django.http.HttpResponse(
        body.encode("utf-8"),
        headers={
            **comm_res.headers,
            **net.create_headers(
                env=client.env,
                framework=FRAMEWORK,
                server_kind=server_kind,
            ),
        },
        status=comm_res.status_code,
    )



import datetime
import hashlib
import inspect
import json
import re
import traceback
import typing

from . import const, errors, types


def get_traceback(err: Exception) -> str:
    return "".join(
        traceback.format_exception(type(err), err, err.__traceback__)
    )


def hash_signing_key(key: str) -> str:
    return hashlib.sha256(
        bytearray.fromhex(remove_signing_key_prefix(key))
    ).hexdigest()


def hash_step_id(step_id: str) -> str:
    return hashlib.sha1(step_id.encode("utf-8")).hexdigest()  # noqa: S324


def dump_json(obj: object) -> types.MaybeError[str]:
    try:
        return json.dumps(obj)
    except Exception as err:
        return errors.OutputUnserializableError(str(err))


def remove_signing_key_prefix(key: str) -> str:
    prefix_match = re.match(r"^signkey-[\w]+-", key)
    prefix = ""
    if prefix_match:
        prefix = prefix_match.group(0)

    return key[len(prefix) :]


def prep_body(obj: types.T) -> types.T:
    """
    Prep body before sending to the Inngest server. This function will:
    - Remove items whose value is None.
    - Convert keys to camelCase.
    """

    if isinstance(obj, dict):
        return {k: prep_body(v) for k, v in obj.items() if v is not None}  # type: ignore
    if isinstance(obj, list):
        return [prep_body(v) for v in obj if v is not None]  # type: ignore
    return obj


class _Duration:
    @classmethod
    def second(cls, count: int = 1) -> int:
        return count * 1000

    @classmethod
    def minute(cls, count: int = 1) -> int:
        return count * cls.second(60)

    @classmethod
    def hour(cls, count: int = 1) -> int:
        return count * cls.minute(60)

    @classmethod
    def day(cls, count: int = 1) -> int:
        return count * cls.hour(24)

    @classmethod
    def week(cls, count: int = 1) -> int:
        return count * cls.day(7)


def to_duration_str(
    ms: typing.Union[int, datetime.timedelta],
) -> types.MaybeError[str]:
    if isinstance(ms, datetime.timedelta):
        ms = int(ms.total_seconds() * 1000)

    if ms < _Duration.second():
        return errors.FunctionConfigInvalidError(
            "duration must be at least 1 second"
        )

    if ms < _Duration.minute():
        return f"{ms // _Duration.second()}s"
    if ms < _Duration.hour():
        return f"{ms // _Duration.minute()}m"
    if ms < _Duration.day():
        return f"{ms // _Duration.hour()}h"
    if ms < _Duration.week():
        return f"{ms // _Duration.day()}d"

    return f"{ms // _Duration.week()}w"


def to_iso_utc(value: datetime.datetime) -> str:
    return (
        value.astimezone(datetime.timezone.utc).strftime(
            "%Y-%m-%dT%H:%M:%S.%f"
        )[:-3]
        + "Z"
    )


def get_server_kind(
    headers: dict[str, str],
) -> typing.Union[const.ServerKind, None, Exception]:
    value = headers.get(const.HeaderKey.SERVER_KIND.value, None)
    if value is None:
        return None

    try:
        return const.ServerKind(value)
    except ValueError:
        return Exception(f"invalid server kind: {value}")


async def maybe_await(
    value: typing.Union[types.T, typing.Awaitable[types.T]],
) -> types.T:
    if inspect.isawaitable(value):
        return await value  # type: ignore

    return value  # type: ignore


def remove_first_traceback_frame(err: Exception) -> None:
    """
    Remove the first frame from the traceback, since we don't want our internal
    code to appear in the traceback.
    """

    if err.__traceback__:
        err.__traceback__ = err.__traceback__.tb_next



import datetime

from . import function_config


def test_serialization() -> None:
    data = function_config.FunctionConfig(
        batch_events=function_config.Batch(
            max_size=10,
            timeout=datetime.timedelta(seconds=60),
        ),
        cancel=[
            function_config.Cancel(
                event="foo",
                if_exp="foo",
                timeout=datetime.timedelta(seconds=60),
            )
        ],
        concurrency=[
            function_config.Concurrency(
                key="foo",
                limit=1,
                scope="account",
            )
        ],
        debounce=function_config.Debounce(
            key="foo",
            period=datetime.timedelta(seconds=60),
        ),
        id="foo",
        name="foo",
        steps={
            "foo": function_config.Step(
                id="foo",
                name="foo",
                retries=function_config.Retries(attempts=1),
                runtime=function_config.Runtime(type="http", url="foo"),
            )
        },
        rate_limit=function_config.RateLimit(
            key="foo",
            limit=1,
            period=datetime.timedelta(seconds=60),
        ),
        throttle=function_config.Throttle(
            key="foo",
            count=1,
            period=datetime.timedelta(seconds=60),
        ),
        triggers=[
            function_config.TriggerCron(cron="foo"),
            function_config.TriggerEvent(event="foo", expression="foo"),
        ],
    ).to_dict()
    if isinstance(data, Exception):
        raise data

    assert data == {
        "batchEvents": {
            "maxSize": 10,
            "timeout": "1m",
        },
        "cancel": [
            {
                "event": "foo",
                "if": "foo",
                "timeout": "1m",
            }
        ],
        "concurrency": [
            {
                "key": "foo",
                "limit": 1,
                "scope": "account",
            }
        ],
        "debounce": {
            "key": "foo",
            "period": "1m",
        },
        "id": "foo",
        "name": "foo",
        "rateLimit": {
            "key": "foo",
            "limit": 1,
            "period": "1m",
        },
        "steps": {
            "foo": {
                "id": "foo",
                "name": "foo",
                "retries": {"attempts": 1},
                "runtime": {
                    "type": "http",
                    "url": "foo",
                },
            }
        },
        "throttle": {
            "key": "foo",
            "count": 1,
            "period": "1m",
        },
        "triggers": [
            {"cron": "foo"},
            {
                "event": "foo",
                "expression": "foo",
            },
        ],
    }



from __future__ import annotations

import datetime
import typing

import pydantic

from . import errors, transforms, types


class _BaseConfig(types.BaseModel):
    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        return errors.FunctionConfigInvalidError.from_validation_error(err)


class Batch(_BaseConfig):
    max_size: int = pydantic.Field(..., serialization_alias="maxSize")
    timeout: typing.Union[int, datetime.timedelta, None] = None

    @pydantic.field_serializer("timeout")
    def serialize_timeout(
        self,
        value: typing.Union[int, datetime.timedelta, None],
    ) -> typing.Optional[str]:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Cancel(_BaseConfig):
    event: str
    if_exp: typing.Optional[str] = pydantic.Field(
        default=None, serialization_alias="if"
    )
    timeout: typing.Union[int, datetime.timedelta, None] = None

    @pydantic.field_serializer("timeout")
    def serialize_timeout(
        self,
        value: typing.Union[int, datetime.timedelta, None],
    ) -> typing.Optional[str]:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Concurrency(_BaseConfig):
    key: typing.Optional[str] = None
    limit: int
    scope: typing.Optional[typing.Literal["account", "env", "fn"]] = None


class Debounce(_BaseConfig):
    key: typing.Optional[str] = None
    period: typing.Union[int, datetime.timedelta]

    @pydantic.field_serializer("period")
    def serialize_period(
        self,
        value: typing.Union[int, datetime.timedelta, None],
    ) -> typing.Optional[str]:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class FunctionConfig(_BaseConfig):
    batch_events: typing.Optional[Batch] = pydantic.Field(
        default=None, serialization_alias="batchEvents"
    )
    cancel: typing.Optional[list[Cancel]] = None
    concurrency: typing.Optional[list[Concurrency]] = None
    debounce: typing.Optional[Debounce] = None
    id: str
    name: typing.Optional[str] = None
    rate_limit: typing.Optional[RateLimit] = pydantic.Field(
        default=None, serialization_alias="rateLimit"
    )
    steps: dict[str, Step]
    throttle: typing.Optional[Throttle] = None
    triggers: list[typing.Union[TriggerCron, TriggerEvent]]


class RateLimit(_BaseConfig):
    key: typing.Optional[str] = None
    limit: int
    period: typing.Union[int, datetime.timedelta]

    @pydantic.field_serializer("period")
    def serialize_period(
        self,
        value: typing.Union[int, datetime.timedelta, None],
    ) -> typing.Optional[str]:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Retries(_BaseConfig):
    attempts: int = pydantic.Field(ge=0)


class Runtime(_BaseConfig):
    type: typing.Literal["http"]
    url: str


class Step(_BaseConfig):
    id: str
    name: str
    retries: typing.Optional[Retries] = None
    runtime: Runtime


class Throttle(_BaseConfig):
    key: typing.Optional[str] = None
    count: int
    period: typing.Union[int, datetime.timedelta]

    @pydantic.field_serializer("period")
    def serialize_period(
        self,
        value: typing.Union[int, datetime.timedelta, None],
    ) -> typing.Optional[str]:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class TriggerCron(_BaseConfig):
    cron: str


class TriggerEvent(_BaseConfig):
    event: str
    expression: typing.Optional[str] = None



from .base import (
    FunctionID,
    ResponseInterrupt,
    SkipInterrupt,
    StepIDCounter,
    StepMemos,
)
from .step_async import Step
from .step_sync import StepSync

__all__ = [
    "FunctionID",
    "ResponseInterrupt",
    "SkipInterrupt",
    "Step",
    "StepIDCounter",
    "StepMemos",
    "StepSync",
]



"""
DigitalOcean integration for Inngest
"""

from __future__ import annotations

import json
import typing
import urllib.parse

from ._internal import (
    client_lib,
    comm,
    const,
    errors,
    execution,
    function,
    net,
    transforms,
    types,
)

FRAMEWORK = const.Framework.DIGITAL_OCEAN


def serve(
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    serve_origin: typing.Optional[str] = None,
    serve_path: typing.Optional[str] = None,
) -> typing.Callable[[dict[str, object], _Context], _Response]:
    """
    Serve Inngest functions in a DigitalOcean Function.

    Args:
    ----
        client: Inngest client.
        functions: List of functions to serve.

        serve_origin: Origin to serve the functions from.
        serve_path: The entire function path (e.g. /api/v1/web/fn-b094417f/sample/hello).
    """

    handler = comm.CommHandler(
        api_base_url=client.api_origin,
        client=client,
        framework=FRAMEWORK,
        functions=functions,
    )

    def main(event: dict[str, object], context: _Context) -> _Response:
        server_kind: typing.Optional[const.ServerKind] = None

        try:
            if not isinstance(event, dict) and "http" not in event:
                raise errors.BodyInvalidError('missing "http" key in event')

            http = _EventHTTP.from_raw(event["http"])
            if isinstance(http, Exception):
                raise errors.BodyInvalidError(http)

            if http.headers is None:
                raise errors.BodyInvalidError(
                    'missing "headers" event.http; have you set "web: raw"?'
                )
            if http.queryString is None:
                raise errors.BodyInvalidError(
                    'missing "queryString" event.http; have you set "web: raw"?'
                )

            headers = net.normalize_headers(http.headers)

            _server_kind = transforms.get_server_kind(headers)
            if not isinstance(_server_kind, Exception):
                server_kind = _server_kind
            else:
                client.logger.error(_server_kind)
                server_kind = None

            req_sig = net.RequestSignature(
                body=_to_body_bytes(http.body),
                headers=headers,
                mode=client._mode,
            )

            query_params = urllib.parse.parse_qs(http.queryString)

            if http.method == "GET":
                return _to_response(
                    client,
                    handler.inspect(server_kind, req_sig),
                    server_kind,
                )

            if http.method == "POST":
                if http.body is None:
                    raise errors.BodyInvalidError(
                        'missing "body" event.http; have you set "web: raw"?'
                    )

                body = json.loads(http.body)
                if not isinstance(body, dict):
                    raise errors.BodyInvalidError("body must be an object")

                # These are sent as query params but DigitalOcean munges them with the body
                fn_id = _get_first(
                    query_params.get(const.QueryParamKey.FUNCTION_ID.value),
                )
                step_id = _get_first(
                    query_params.get(const.QueryParamKey.STEP_ID.value),
                )

                if fn_id is None:
                    raise errors.QueryParamMissingError(
                        const.QueryParamKey.FUNCTION_ID.value
                    )
                if step_id is None:
                    raise errors.QueryParamMissingError(
                        const.QueryParamKey.STEP_ID.value
                    )
                call = execution.Call.from_raw(body)
                if isinstance(call, Exception):
                    raise call

                return _to_response(
                    client,
                    handler.call_function_sync(
                        call=call,
                        fn_id=fn_id,
                        req_sig=req_sig,
                        target_hashed_id=step_id,
                    ),
                    server_kind,
                )

            if http.method == "PUT":
                # DigitalOcean does not give the full path to the function, so
                # we'll build it by hardcoding the path prefix ("api/v1/web")
                # and concatenating it with the function name. This should be
                # identical to the path, but DigitalOcean may change this in the
                # future (e.g. a new API version).
                #
                # You might be tempted to use event.http.path, but that's
                # actually the relative path after the prefix + function name.
                path = "/api/v1/web" + context.function_name

                request_url = urllib.parse.urljoin(context.api_host, path)
                sync_id = _get_first(
                    query_params.get(const.QueryParamKey.SYNC_ID.value),
                )

                return _to_response(
                    client,
                    handler.register_sync(
                        app_url=net.create_serve_url(
                            request_url=request_url,
                            serve_origin=serve_origin,
                            serve_path=serve_path,
                        ),
                        server_kind=server_kind,
                        sync_id=sync_id,
                    ),
                    server_kind,
                )

            raise Exception(f"unsupported method: {http.method}")
        except Exception as e:
            comm_res = comm.CommResponse.from_error(client.logger, e)
            if isinstance(
                e, (errors.BodyInvalidError, errors.QueryParamMissingError)
            ):
                comm_res.status_code = 400

            return _to_response(
                client,
                comm_res,
                server_kind,
            )

    return main


def _get_first(
    items: typing.Optional[list[types.T]],
) -> typing.Optional[types.T]:
    if items is None or len(items) == 0:
        return None
    return items[0]


def _to_body_bytes(body: typing.Optional[str]) -> bytes:
    if body is None:
        return b""
    return body.encode("utf-8")


def _to_response(
    client: client_lib.Inngest,
    comm_res: comm.CommResponse,
    server_kind: typing.Union[const.ServerKind, None],
) -> _Response:
    return {
        "body": comm_res.body,  # type: ignore
        "headers": {
            **comm_res.headers,
            **net.create_headers(
                env=client.env,
                framework=FRAMEWORK,
                server_kind=server_kind,
            ),
        },
        "statusCode": comm_res.status_code,
    }


class _EventHTTP(types.BaseModel):
    body: typing.Optional[str] = None
    headers: typing.Optional[dict[str, str]] = None
    method: typing.Optional[str] = None
    path: typing.Optional[str] = None
    queryString: typing.Optional[str] = None  # noqa: N815


class _Context(typing.Protocol):
    # E.g. "https://faas-nyc1-2ef2e6cc.doserverless.co"
    api_host: str

    # E.g. "/fn-b094417f/sample/hello"
    function_name: str


class _Response(typing.TypedDict):
    body: str
    headers: dict[str, str]
    statusCode: int



from __future__ import annotations

import typing

import pydantic

from . import const, transforms


class Error(Exception):
    """
    Base error for all our custom errors
    """

    code: const.ErrorCode
    include_stack: bool = True
    is_retriable: bool = True

    @property
    def message(self) -> str:
        return str(self)

    @property
    def name(self) -> str:
        return type(self).__name__

    @property
    def stack(self) -> typing.Optional[str]:
        if self.include_stack is False:
            return None

        return transforms.get_traceback(self)


class URLInvalidError(Error):
    code = const.ErrorCode.URL_INVALID


class FunctionConfigInvalidError(Error):
    code = const.ErrorCode.FUNCTION_CONFIG_INVALID

    @classmethod
    def from_validation_error(
        cls,
        err: pydantic.ValidationError,
    ) -> FunctionConfigInvalidError:
        """
        Extract info from Pydantic's ValidationError and return our internal
        InvalidFunctionConfig error.
        """
        default = cls(str(err))

        errors = err.errors()
        if len(errors) == 0:
            return default
        loc = errors[0].get("loc")
        if loc is None or len(loc) == 0:
            return default

        field = ""
        for part in loc:
            if isinstance(part, int):
                return default
            if len(field) > 0:
                field += "."
            field += part

        msg = errors[0].get("msg")
        if msg is None:
            return default

        return cls(f"{loc[0]}: {msg}")


class AsyncUnsupportedError(Error):
    code = const.ErrorCode.ASYNC_UNSUPPORTED


class SigVerificationFailedError(Error):
    code = const.ErrorCode.SIG_VERIFICATION_FAILED


class BodyInvalidError(Error):
    code = const.ErrorCode.BODY_INVALID


class EventKeyUnspecifiedError(Error):
    code = const.ErrorCode.EVENT_KEY_UNSPECIFIED


class FunctionNotFoundError(Error):
    code = const.ErrorCode.FUNCTION_NOT_FOUND


class HeaderMissingError(Error):
    code = const.ErrorCode.HEADER_MISSING


class QueryParamMissingError(Error):
    code = const.ErrorCode.QUERY_PARAM_MISSING


class SigningKeyMissingError(Error):
    code = const.ErrorCode.SIGNING_KEY_UNSPECIFIED


class RegistrationFailedError(Error):
    code = const.ErrorCode.REGISTRATION_FAILED


class UnknownError(Error):
    code = const.ErrorCode.UNKNOWN


class StepUnexpectedError(Error):
    code = const.ErrorCode.STEP_UNEXPECTED
    include_stack = False


class OutputUnserializableError(Error):
    code = const.ErrorCode.OUTPUT_UNSERIALIZABLE


class NonRetriableError(Error):
    """End users can raise this error to prevent retries."""

    code = const.ErrorCode.NON_RETRIABLE_ERROR
    is_retriable = False

    def __init__(
        self,
        message: typing.Optional[str] = None,
        cause: typing.Optional[typing.Mapping[str, object]] = None,
    ) -> None:
        super().__init__(message)
        self.cause = cause


class StepError(Error):
    """
    Wraps a userland error. This is necessary because the Executor sends
    memoized error data which can't be deserialized into the original error
    class.
    """

    code = const.ErrorCode.STEP_ERRORED

    # Not retriable since this error is thrown after exhausting retries
    is_retriable = False

    @property
    def message(self) -> str:
        """
        Returns the userland error message
        """

        return self._message

    @property
    def name(self) -> str:
        """
        Returns the userland error name
        """

        return self._name

    @property
    def stack(self) -> typing.Optional[str]:
        """
        Returns the userland error stack trace
        """

        return self._stack

    def __init__(
        self,
        message: str,
        name: str,
        stack: typing.Optional[str],
    ) -> None:
        """
        Args:
        ----
            message: Userland error's message
            name: Userland error's name
            stack: Userland error's stack trace
        """

        super().__init__(message)
        self._message = message
        self._name = name
        self._stack = stack



from __future__ import annotations

import dataclasses
import threading
import typing

import pydantic

from inngest._internal import (
    client_lib,
    errors,
    execution,
    middleware_lib,
    transforms,
    types,
)


class StepMemos:
    """Holds memoized step output."""

    @property
    def size(self) -> int:
        return len(self._memos)

    def __init__(self, memos: dict[str, execution.Output]) -> None:
        self._memos = memos

    def values(self) -> typing.Iterator[execution.Output]:
        return iter(self._memos.values())

    def pop(
        self, hashed_id: str
    ) -> typing.Union[execution.Output, types.EmptySentinel]:
        if hashed_id in self._memos:
            memo = self._memos[hashed_id]

            # Remove memo
            self._memos = {
                k: v for k, v in self._memos.items() if k != hashed_id
            }

            return memo

        return types.empty_sentinel

    @classmethod
    def from_raw(cls, raw: dict[str, object]) -> StepMemos:
        memos = {}
        for k, v in raw.items():
            output = execution.Output.from_raw(v)
            if isinstance(output, Exception):
                # Not all steps nest their output in an Output-compatible object
                # (i.e. they don't nest output in a data field). For example,
                # `step.run` nests its output in a data field but
                # `step.waitForEvent` will not nest its fulfilling event.
                output = execution.Output(data=v)

            memos[k] = output
        return cls(memos)


class StepBase:
    def __init__(
        self,
        client: client_lib.Inngest,
        memos: StepMemos,
        middleware: middleware_lib.MiddlewareManager,
        step_id_counter: StepIDCounter,
        target_hashed_id: typing.Optional[str],
    ) -> None:
        self._client = client
        self._inside_parallel = False
        self._memos = memos
        self._middleware = middleware
        self._step_id_counter = step_id_counter
        self._target_hashed_id = target_hashed_id

    async def _get_memo(
        self,
        hashed_id: str,
    ) -> typing.Union[execution.Output, types.EmptySentinel]:
        memo = self._memos.pop(hashed_id)

        # If there are no more memos then all future code is new.
        if self._memos.size == 0:
            await self._middleware.before_execution()

        if not isinstance(memo, types.EmptySentinel):
            if memo.error is not None:
                error = execution.MemoizedError.from_raw(memo.error)
                if isinstance(error, Exception):
                    raise error

                # If there's a memoized error then raise an error, since the
                # step exhausted its retries
                raise errors.StepError(
                    message=error.message,
                    name=error.name,
                    stack=error.stack,
                )

        return memo

    def _get_memo_sync(
        self,
        hashed_id: str,
    ) -> typing.Union[execution.Output, types.EmptySentinel]:
        memo = self._memos.pop(hashed_id)

        # If there are no more memos then all future code is new.
        if self._memos.size == 0:
            self._middleware.before_execution_sync()

        if not isinstance(memo, types.EmptySentinel):
            if memo.error is not None:
                error = execution.MemoizedError.from_raw(memo.error)
                if isinstance(error, Exception):
                    raise error

                # If there's a memoized error then raise an error, since the
                # step exhausted its retries
                raise errors.StepError(
                    message=error.message,
                    name=error.name,
                    stack=error.stack,
                )

        return memo

    def _handle_skip(
        self,
        parsed_step_id: _ParsedStepID,
    ) -> None:
        """
        Handle a skip interrupt. Step targeting is enabled and this step is not
        the target then skip the step.
        """

        is_targeting_enabled = self._target_hashed_id is not None
        is_targeted = self._target_hashed_id == parsed_step_id.hashed
        if is_targeting_enabled and not is_targeted:
            # Skip this step because a different step is targeted.
            raise SkipInterrupt(parsed_step_id.user_facing)

    def _parse_step_id(self, step_id: str) -> _ParsedStepID:
        """
        Parse a user-specified step ID into a hashed ID and a deduped
        user-facing step ID.
        """

        id_count = self._step_id_counter.increment(step_id)
        if id_count > 1:
            step_id = f"{step_id}:{id_count - 1}"

        return _ParsedStepID(
            hashed=transforms.hash_step_id(step_id),
            user_facing=step_id,
        )


@dataclasses.dataclass
class _ParsedStepID:
    hashed: str
    user_facing: str


class StepIDCounter:
    """
    Counts the number of times a step ID has been used. We support reused step
    IDs so we need a way to keep track of their counts.
    """

    def __init__(self) -> None:
        self._counts: dict[str, int] = {}
        self._mutex = threading.Lock()

    def increment(self, hashed_id: str) -> int:
        with self._mutex:
            if hashed_id not in self._counts:
                self._counts[hashed_id] = 0

            self._counts[hashed_id] += 1
            return self._counts[hashed_id]


class ResponseInterrupt(BaseException):
    """
    Extend BaseException to avoid being caught by the user's code. Users can
    still catch it if they do a "bare except", but that's a known antipattern in
    the Python world.
    """

    def __init__(
        self,
        responses: typing.Union[
            execution.StepResponse, list[execution.StepResponse]
        ],
    ) -> None:
        if not isinstance(responses, list):
            responses = [responses]
        self.responses = responses


class SkipInterrupt(BaseException):
    def __init__(self, step_id: str) -> None:
        self.step_id = step_id


@dataclasses.dataclass
class FunctionID:
    app_id: str
    function_id: str


class InvokeOpts(types.BaseModel):
    function_id: str
    payload: InvokeOptsPayload


class InvokeOptsPayload(types.BaseModel):
    data: object
    user: object
    v: typing.Optional[str]


class WaitForEventOpts(types.BaseModel):
    if_exp: typing.Optional[str] = pydantic.Field(..., serialization_alias="if")
    timeout: str



from __future__ import annotations

import asyncio
import http
import os
import typing
import urllib.parse

import httpx

from inngest._internal import (
    client_lib,
    const,
    errors,
    execution,
    function,
    function_config,
    middleware_lib,
    net,
    registration,
    step_lib,
    transforms,
    types,
)


class CommResponse:
    def __init__(
        self,
        *,
        body: object = None,
        headers: typing.Optional[dict[str, str]] = None,
        status_code: int = http.HTTPStatus.OK.value,
    ) -> None:
        self.headers = headers or {}
        self.body = body
        self.status_code = status_code

    @classmethod
    def from_call_result(
        cls,
        logger: types.Logger,
        call_res: execution.CallResult,
    ) -> CommResponse:
        headers = {
            const.HeaderKey.SERVER_TIMING.value: "handler",
        }

        if execution.is_step_call_responses(call_res):
            out: list[dict[str, object]] = []
            for item in call_res:
                d = item.to_dict()
                if isinstance(d, Exception):
                    return cls.from_error(
                        logger,
                        errors.OutputUnserializableError(
                            f'"{item.display_name}" returned unserializable data'
                        ),
                    )

                # Unnest data and error fields to work with the StepRun opcode.
                # They should probably be unnested lower in the code, but this
                # is a quick fix that doesn't break middleware contracts
                nested_data = d.get("data")
                if isinstance(nested_data, dict):
                    d["data"] = nested_data.get("data")
                    d["error"] = nested_data.get("error")

                out.append(d)

            return cls(
                body=transforms.prep_body(out),
                headers=headers,
                status_code=http.HTTPStatus.PARTIAL_CONTENT.value,
            )

        if isinstance(call_res, execution.CallError):
            logger.error(call_res.stack)

            d = call_res.to_dict()
            if isinstance(d, Exception):
                return cls.from_error(logger, d)

            if call_res.is_retriable is False:
                headers[const.HeaderKey.NO_RETRY.value] = "true"

            return cls(
                body=transforms.prep_body(d),
                headers=headers,
                status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR.value,
            )

        if isinstance(call_res, execution.FunctionCallResponse):
            return cls(
                body=call_res.data,
                headers=headers,
            )

        return cls.from_error(
            logger,
            errors.UnknownError("unknown call result"),
        )

    @classmethod
    def from_error(
        cls,
        logger: types.Logger,
        err: Exception,
        status: http.HTTPStatus = http.HTTPStatus.INTERNAL_SERVER_ERROR,
    ) -> CommResponse:
        code: typing.Optional[str] = None
        if isinstance(err, errors.Error):
            code = err.code.value
        else:
            code = const.ErrorCode.UNKNOWN.value

        logger.error(f"{code}: {err!s}")

        return cls(
            body={
                "code": code,
                "message": str(err),
                "name": type(err).__name__,
            },
            status_code=status.value,
        )

    @classmethod
    def from_error_code(
        cls,
        code: const.ErrorCode,
        message: str,
        status: http.HTTPStatus = http.HTTPStatus.INTERNAL_SERVER_ERROR,
    ) -> CommResponse:
        return cls(
            body={
                "code": code.value,
                "message": message,
            },
            status_code=status.value,
        )


class CommHandler:
    _base_url: str
    _client: client_lib.Inngest
    _fns: dict[str, function.Function]
    _framework: const.Framework
    _mode: const.ServerKind
    _signing_key: typing.Optional[str]
    _signing_key_fallback: typing.Optional[str]

    def __init__(
        self,
        *,
        api_base_url: typing.Optional[str] = None,
        client: client_lib.Inngest,
        framework: const.Framework,
        functions: list[function.Function],
    ) -> None:
        self._client = client

        self._mode = client._mode

        api_base_url = api_base_url or os.getenv(
            const.EnvKey.API_BASE_URL.value
        )
        if api_base_url is None:
            if self._mode == const.ServerKind.DEV_SERVER:
                api_base_url = const.DEV_SERVER_ORIGIN
            else:
                api_base_url = const.DEFAULT_API_ORIGIN

        try:
            self._api_origin = net.parse_url(api_base_url)
        except Exception as err:
            raise errors.URLInvalidError() from err

        self._fns = {fn.get_id(): fn for fn in functions}
        self._framework = framework

        signing_key = client.signing_key
        if signing_key is None:
            if self._client.is_production:
                signing_key = os.getenv(const.EnvKey.SIGNING_KEY.value)
                if signing_key is None:
                    self._client.logger.error("missing signing key")
                    raise errors.SigningKeyMissingError()
        self._signing_key = signing_key

        self._signing_key_fallback = client.signing_key_fallback

    def _build_registration_request(
        self,
        *,
        app_url: str,
        server_kind: typing.Optional[const.ServerKind],
        sync_id: typing.Optional[str],
    ) -> types.MaybeError[httpx.Request]:
        registration_url = urllib.parse.urljoin(
            self._api_origin,
            "/fn/register",
        )

        fn_configs = self.get_function_configs(app_url)
        if isinstance(fn_configs, Exception):
            return fn_configs

        body = registration.RegisterRequest(
            app_name=self._client.app_id,
            deploy_type=registration.DeployType.PING,
            framework=self._framework,
            functions=fn_configs,
            sdk=f"{const.LANGUAGE}:v{const.VERSION}",
            url=app_url,
            v="0.1",
        ).to_dict()
        if isinstance(body, Exception):
            return body

        headers = net.create_headers(
            env=self._client.env,
            framework=self._framework,
            server_kind=server_kind,
        )

        params = {}
        if sync_id is not None:
            params[const.QueryParamKey.SYNC_ID.value] = sync_id

        return self._client._http_client_sync.build_request(
            "POST",
            registration_url,
            headers=headers,
            json=transforms.prep_body(body),
            params=params,
            timeout=30,
        )

    async def call_function(
        self,
        *,
        call: execution.Call,
        fn_id: str,
        req_sig: net.RequestSignature,
        target_hashed_id: str,
    ) -> CommResponse:
        """Handle a function call from the Executor."""

        if target_hashed_id == execution.UNSPECIFIED_STEP_ID:
            target_step_id = None
        else:
            target_step_id = target_hashed_id

        middleware = middleware_lib.MiddlewareManager.from_client(self._client)

        # Validate the request signature.
        err = req_sig.validate(
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(err, Exception):
            return await self._respond(middleware, err)

        # Get the function we should call.
        fn = self._get_function(fn_id)
        if isinstance(fn, Exception):
            return await self._respond(middleware, fn)

        events = call.events
        steps = call.steps
        if call.use_api:
            # Putting the batch and memoized steps in the request would make it
            # to big, so the Executor is telling the SDK to fetch them from the
            # API

            try:
                events, steps = await asyncio.gather(
                    self._client._get_batch(call.ctx.run_id),
                    self._client._get_steps(call.ctx.run_id),
                )
            except Exception as err:
                return await self._respond(middleware, err)
        if events is None:
            # Should be unreachable. The Executor should always either send the
            # batch or tell the SDK to fetch the batch

            return await self._respond(
                middleware, Exception("events not in request")
            )

        call_res = await fn.call(
            self._client,
            function.Context(
                _steps=step_lib.StepMemos.from_raw(steps),
                attempt=call.ctx.attempt,
                event=call.event,
                events=events,
                logger=self._client.logger,
                run_id=call.ctx.run_id,
            ),
            fn_id,
            middleware,
            target_step_id,
        )

        return await self._respond(middleware, call_res)

    def call_function_sync(
        self,
        *,
        call: execution.Call,
        fn_id: str,
        req_sig: net.RequestSignature,
        target_hashed_id: str,
    ) -> CommResponse:
        """Handle a function call from the Executor."""

        if target_hashed_id == execution.UNSPECIFIED_STEP_ID:
            target_step_id = None
        else:
            target_step_id = target_hashed_id

        middleware = middleware_lib.MiddlewareManager.from_client(self._client)

        # Validate the request signature.
        err = req_sig.validate(
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(err, Exception):
            return self._respond_sync(middleware, err)

        # Get the function we should call.
        fn = self._get_function(fn_id)
        if isinstance(fn, Exception):
            return self._respond_sync(middleware, fn)

        events = call.events
        steps = call.steps
        if call.use_api:
            # Putting the batch and memoized steps in the request would make it
            # to big, so the Executor is telling the SDK to fetch them from the
            # API

            try:
                events = self._client._get_batch_sync(call.ctx.run_id)
                steps = self._client._get_steps_sync(call.ctx.run_id)
            except Exception as err:
                return self._respond_sync(middleware, err)
        if events is None:
            # Should be unreachable. The Executor should always either send the
            # batch or tell the SDK to fetch the batch

            return self._respond_sync(
                middleware, Exception("events not in request")
            )

        call_res = fn.call_sync(
            self._client,
            function.Context(
                _steps=step_lib.StepMemos.from_raw(steps),
                attempt=call.ctx.attempt,
                event=call.event,
                events=events,
                logger=self._client.logger,
                run_id=call.ctx.run_id,
            ),
            fn_id,
            middleware,
            target_step_id,
        )

        return self._respond_sync(middleware, call_res)

    def _get_function(self, fn_id: str) -> types.MaybeError[function.Function]:
        # Look for the function ID in the list of user functions, but also
        # look for it in the list of on_failure functions.
        for _fn in self._fns.values():
            if _fn.get_id() == fn_id:
                return _fn
            if _fn.on_failure_fn_id == fn_id:
                return _fn

        # If we didn't find the function ID, it might be because the function ID
        # in the request uses the old format that didn't include the app ID.
        # We'll prefix the function ID with the app ID and try again. This logic
        # can be deleted when no one is using Python SDK versions below 0.3.0
        # anymore.
        app_and_fn_id = f"{self._client.app_id}-{fn_id}"
        for _fn in self._fns.values():
            if _fn.get_id() == app_and_fn_id:
                return _fn
            if _fn.on_failure_fn_id == app_and_fn_id:
                return _fn

        return errors.FunctionNotFoundError(f"function {fn_id} not found")

    def get_function_configs(
        self,
        app_url: str,
    ) -> types.MaybeError[list[function_config.FunctionConfig]]:
        configs: list[function_config.FunctionConfig] = []
        for fn in self._fns.values():
            config = fn.get_config(app_url)
            configs.append(config.main)

            if config.on_failure is not None:
                configs.append(config.on_failure)

        if len(configs) == 0:
            return errors.FunctionConfigInvalidError("no functions found")
        return configs

    def inspect(
        self,
        server_kind: typing.Optional[const.ServerKind],
        req_sig: net.RequestSignature,
    ) -> CommResponse:
        """Handle Dev Server's auto-discovery."""

        if server_kind is not None and server_kind != self._mode:
            # Tell Dev Server to leave the app alone since it's in production
            # mode.
            return CommResponse(
                body={},
                headers={},
                status_code=403,
            )

        # Validate the request signature.
        err = req_sig.validate(
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if self._client._mode != const.ServerKind.CLOUD or isinstance(
            err, Exception
        ):
            body = _InsecureInspection(
                function_count=len(self._fns),
                has_event_key=self._client.event_key is not None,
                has_signing_key=self._signing_key is not None,
                mode=self._mode,
            )
        else:
            signing_key_hash = (
                transforms.hash_signing_key(self._signing_key)
                if self._signing_key
                else None
            )

            signing_key_fallback_hash = (
                transforms.hash_signing_key(self._signing_key_fallback)
                if self._signing_key_fallback
                else None
            )

            body = _SecureInspection(
                function_count=len(self._fns),
                has_event_key=self._client.event_key is not None,
                has_signing_key=self._signing_key is not None,
                mode=self._mode,
                signing_key_fallback_hash=signing_key_fallback_hash,
                signing_key_hash=signing_key_hash,
            )

        body_json = body.to_dict()
        if isinstance(body, Exception):
            body_json = {
                "error": "failed to serialize inspection data",
            }

        return CommResponse(
            body=body_json,
            headers=net.create_headers(
                env=self._client.env,
                framework=self._framework,
                server_kind=server_kind,
            ),
            status_code=200,
        )

    def _parse_registration_response(
        self,
        server_res: httpx.Response,
        server_kind: typing.Optional[const.ServerKind],
    ) -> CommResponse:
        try:
            server_res_body = server_res.json()
        except Exception:
            return CommResponse.from_error(
                self._client.logger,
                errors.RegistrationFailedError("response is not valid JSON"),
            )

        if not isinstance(server_res_body, dict):
            return CommResponse.from_error(
                self._client.logger,
                errors.RegistrationFailedError("response is not an object"),
            )

        if server_res.status_code < 400:
            return CommResponse(
                body=server_res_body,
                headers=net.create_headers(
                    env=self._client.env,
                    framework=self._framework,
                    server_kind=server_kind,
                ),
                status_code=http.HTTPStatus.OK,
            )

        msg = server_res_body.get("error")
        if not isinstance(msg, str):
            msg = "registration failed"
        comm_res = CommResponse.from_error(
            self._client.logger,
            errors.RegistrationFailedError(msg.strip()),
        )
        comm_res.status_code = server_res.status_code
        return comm_res

    async def register(
        self,
        *,
        app_url: str,
        server_kind: typing.Optional[const.ServerKind],
        sync_id: typing.Optional[str],
    ) -> CommResponse:
        """Handle a registration call."""

        comm_res = self._validate_registration(server_kind)
        if comm_res is not None:
            return comm_res

        req = self._build_registration_request(
            app_url=app_url,
            server_kind=server_kind,
            sync_id=sync_id,
        )
        if isinstance(req, Exception):
            return CommResponse.from_error(self._client.logger, req)

        res = await net.fetch_with_auth_fallback(
            self._client.logger,
            self._client._http_client,
            self._client._http_client_sync,
            req,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )

        return self._parse_registration_response(
            res,
            server_kind,
        )

    def register_sync(
        self,
        *,
        app_url: str,
        server_kind: typing.Optional[const.ServerKind],
        sync_id: typing.Optional[str],
    ) -> CommResponse:
        """Handle a registration call."""

        comm_res = self._validate_registration(server_kind)
        if comm_res is not None:
            return comm_res

        req = self._build_registration_request(
            app_url=app_url,
            server_kind=server_kind,
            sync_id=sync_id,
        )
        if isinstance(req, Exception):
            return CommResponse.from_error(self._client.logger, req)

        res = net.fetch_with_auth_fallback_sync(
            self._client._http_client_sync,
            req,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )

        return self._parse_registration_response(
            res,
            server_kind,
        )

    async def _respond(
        self,
        middleware: middleware_lib.MiddlewareManager,
        value: typing.Union[execution.CallResult, Exception],
    ) -> CommResponse:
        err = await middleware.before_response()
        if isinstance(err, Exception):
            return CommResponse.from_error(self._client.logger, err)

        if isinstance(value, Exception):
            return CommResponse.from_error(self._client.logger, value)

        return CommResponse.from_call_result(self._client.logger, value)

    def _respond_sync(
        self,
        middleware: middleware_lib.MiddlewareManager,
        value: typing.Union[execution.CallResult, Exception],
    ) -> CommResponse:
        err = middleware.before_response_sync()
        if isinstance(err, Exception):
            return CommResponse.from_error(self._client.logger, err)

        if isinstance(value, Exception):
            return CommResponse.from_error(self._client.logger, value)

        return CommResponse.from_call_result(self._client.logger, value)

    def _validate_registration(
        self,
        server_kind: typing.Optional[const.ServerKind],
    ) -> typing.Optional[CommResponse]:
        if server_kind is not None and server_kind != self._mode:
            msg: str
            if server_kind == const.ServerKind.DEV_SERVER:
                msg = "Sync rejected since it's from a Dev Server but expected Cloud"
            else:
                msg = "Sync rejected since it's from Cloud but expected Dev Server"

            self._client.logger.error(msg)
            return CommResponse.from_error_code(
                const.ErrorCode.SERVER_KIND_MISMATCH,
                msg,
                http.HTTPStatus.BAD_REQUEST,
            )

        return None


class _InsecureInspection(types.BaseModel):
    function_count: int
    has_event_key: bool
    has_signing_key: bool
    mode: const.ServerKind


class _SecureInspection(_InsecureInspection):
    signing_key_fallback_hash: typing.Optional[str]
    signing_key_hash: typing.Optional[str]
